"""
Execution module to perform CRUD operations for NSX-T's Tier 0 Gateway
"""
import logging

from salt.exceptions import SaltInvocationError
from saltext.vmware.utils.nsxt_policy_base_resource import NSXTPolicyBaseResource
from saltext.vmware.utils.nsxt_resource_urls import DHCP_RELAY_CONFIG_URL
from saltext.vmware.utils.nsxt_resource_urls import EDGE_CLUSTER_URL
from saltext.vmware.utils.nsxt_resource_urls import EDGE_NODE_URL
from saltext.vmware.utils.nsxt_resource_urls import IPV6_DAD_PROFILE_URL
from saltext.vmware.utils.nsxt_resource_urls import IPV6_NDRA_PROFILE_URL
from saltext.vmware.utils.nsxt_resource_urls import SEGMENT_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_BFD_PEERS
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_BGP_NEIGHBOR_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_LOCALE_SERVICE_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_LS_INTERFACE_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_STATIC_ROUTE_URL

log = logging.getLogger(__name__)


def __virtual__():
    return "nsxt_policy_tier0"


class NSXTTier0(NSXTPolicyBaseResource):
    @classmethod
    def get_spec_identifier(cls):
        return "tier0"

    @staticmethod
    def get_resource_base_url(baseline_args=None):
        return "/infra/tier-0s"

    @staticmethod
    def get_resource_base_query_params():
        return (
            "cursor",
            "include_mark_for_delete_objects",
            "included_fields",
            "page_size",
            "sort_ascending",
            "sort_by",
        )

    def update_resource_params(self, **kwargs):
        self.multi_resource_params = []
        fields = {
            "id",
            "display_name",
            "description",
            "tags",
            "ha_mode",
            "internal_transit_subnets",
            "transit_subnets",
            "failover_mode",
            "rd_admin_field",
            "arp_limit",
            "force_whitelisting",
            "default_rule_logging",
            "disable_firewall",
            "advanced_config",
            "intersite_config",
            "state",
            "_revision",
        }
        resource_params = {}
        for field in fields:
            val = kwargs.get(field)
            if val:
                resource_params[field] = val
        resource_params["resource_type"] = "Tier0"

        resource_params.setdefault("id", resource_params["display_name"])

        ipv6_profile_paths = []

        ipv6_ndra_profile_id = kwargs.get("ipv6_ndra_profile_id")
        if ipv6_ndra_profile_id:
            ipv6_profile_paths.append(IPV6_NDRA_PROFILE_URL + "/" + ipv6_ndra_profile_id)
        ipv6_dad_profile_id = kwargs.get("ipv6_dad_profile_id")
        if ipv6_dad_profile_id:
            ipv6_profile_paths.append(IPV6_DAD_PROFILE_URL + "/" + ipv6_dad_profile_id)
        if ipv6_profile_paths:
            resource_params["ipv6_profile_paths"] = ipv6_profile_paths
        dhcp_config_id = kwargs.get("dhcp_config_id")
        if dhcp_config_id:
            resource_params["dhcp_config_paths"] = [DHCP_RELAY_CONFIG_URL + "/" + dhcp_config_id]
        vrf_config = kwargs.get("vrf_config")
        if vrf_config:
            vrf_resource_params = {}
            tier0_id = vrf_config.get("tier0_id")
            if not tier0_id:
                raise SaltInvocationError(
                    {
                        "resourceType": "vrf_config",
                        "error": "Please specify the ID of the Tier 0 in the vrf_config",
                    }
                )
            vrf_resource_params["tier0_path"] = NSXTTier0.get_resource_base_url() + "/" + tier0_id
            vrf_fields = {"evpn_l2_vni_config", "evpn_transit_vni", "route_distinguisher"}
            # This block can be refactored
            for field in vrf_fields:
                val = vrf_config.get(field)
                if val:
                    vrf_resource_params[field] = val
            if "route_targets" in vrf_config:
                route_targets = vrf_config["route_targets"] or []
                for route_target in route_targets:
                    route_target["resource_type"] = "VrfRouteTargets"
                vrf_resource_params["route_targets"] = route_targets
            resource_params["vrf_config"] = vrf_resource_params
        self.multi_resource_params.append(resource_params)

    def update_parent_info(self, parent_info):
        parent_info["tier0_id"] = self.resource_params.get("id")

    class NSXTTier0StaticRoutes(NSXTPolicyBaseResource):
        @staticmethod
        def get_resource_update_priority():
            # Create this first
            return 2

        def get_spec_identifier(self):
            return NSXTTier0.NSXTTier0StaticRoutes.get_spec_identifier()

        @classmethod
        def get_spec_identifier(cls):
            return "static_routes"

        def update_resource_params(self, **kwargs):
            self.multi_resource_params = []
            fields = {
                "id",
                "display_name",
                "description",
                "enabled_on_secondary",
                "network",
                "next_hops",
                "tags",
                "state",
                "_revision",
            }
            static_routes = kwargs.get("static_routes") or {}

            for static_route in static_routes:
                resource_params = {}
                # This block can be refactored
                for key in fields:
                    val = static_route.get(key)
                    if val:
                        resource_params[key] = val
                if not resource_params.get("id"):
                    # In case of default display name set default can be used and refactor can be done.
                    resource_params["id"] = resource_params["display_name"]
                self.multi_resource_params.append(resource_params)

        @staticmethod
        def get_resource_base_url(parent_info):
            tier0_id = parent_info.get("tier0_id", "default")
            return TIER_0_STATIC_ROUTE_URL.format(tier0_id)

        def update_parent_info(self, parent_info):
            parent_info["static_routes_id"] = self.resource_params["id"]

    class NSXTTier0SRBFDPeer(NSXTPolicyBaseResource):
        def get_spec_identifier(self):
            return NSXTTier0.NSXTTier0StaticRoutes.NSXTTier0SRVFDPeer.get_spec_identifier()

        @classmethod
        def get_spec_identifier(cls):
            return "bfd_peers"

        @staticmethod
        def get_resource_base_url(parent_info):
            tier0_id = parent_info.get("tier0_id", "default")
            return TIER_0_BFD_PEERS.format(tier0_id)

        def update_resource_params(self, **kwargs):
            fields = {
                "id",
                "display_name",
                "description",
                "enabled",
                "scope",
                "source_addresses",
                "tags",
                "peer_address",
                "state",
                "_revision",
            }
            self.multi_resource_params = []
            bfd_peers = kwargs.get("bfd_peers") or {}
            for bfd_peer in bfd_peers:
                resource_params = {}
                # This block can be refactored
                for key in fields:
                    if bfd_peer.get(key):
                        resource_params[key] = bfd_peer.get(key)
                bfd_profile_id = bfd_peer.get("bfd_profile_id")
                if bfd_profile_id:
                    resource_params["bfd_profile_path"] = "/infra/bfd-profiles/{}".format(
                        bfd_profile_id
                    )
                if not "id" in bfd_peer:
                    resource_params["id"] = resource_params["display_name"]
                resource_params["resource_type"] = "StaticRouteBfdPeer"
                self.multi_resource_params.append(resource_params)

    class NSXTTier0LocaleService(NSXTPolicyBaseResource):
        def get_spec_identifier(self):
            return NSXTTier0.NSXTTier0LocaleService.get_spec_identifier()

        @classmethod
        def get_spec_identifier(cls):
            return "locale_services"

        @staticmethod
        def get_resource_base_url(parent_info):
            tier0_id = parent_info.get("tier0_id", "default")
            return TIER_0_LOCALE_SERVICE_URL.format(tier0_id)

        def update_resource_params(self, **kwargs):
            self.multi_resource_params = []
            fields = {
                "tags",
                "route_redistribution_config",
                "id",
                "display_name",
                "state",
                "description",
                "_revision",
            }
            locale_services = kwargs.get("locale_services") or {}
            for locale_service in locale_services:
                resource_params = {}
                # This block can be refactored
                for field in fields:
                    if locale_service.get(field):
                        resource_params[field] = locale_service[field]
                resource_params["resource_type"] = "LocaleServices"
                edge_cluster_info = locale_service.get("edge_cluster_info")
                if edge_cluster_info:
                    site_id = edge_cluster_info["site_id"]
                    enforcementpoint_id = edge_cluster_info["enforcementpoint_id"]
                    edge_cluster_base_url = EDGE_CLUSTER_URL.format(site_id, enforcementpoint_id)
                    edge_cluster_id = edge_cluster_info.get("edge_cluster_id")
                    resource_params["edge_cluster_path"] = (
                        edge_cluster_base_url + "/" + edge_cluster_id
                    )
                preferred_edge_nodes_info = locale_service.get("preferred_edge_nodes_info")
                if preferred_edge_nodes_info:
                    resource_params["preferred_edge_paths"] = []
                    for preferred_edge_node_info in preferred_edge_nodes_info:
                        site_id = preferred_edge_node_info.get("site_id", "default")
                        enforcementpoint_id = preferred_edge_node_info.get(
                            "enforcementpoint_id", "default"
                        )
                        edge_cluster_id = preferred_edge_node_info.get("edge_cluster_id")
                        edge_node_base_url = EDGE_NODE_URL.format(
                            site_id, enforcementpoint_id, edge_cluster_id
                        )
                        edge_node_id = preferred_edge_node_info.get("edge_node_id")
                        resource_params["preferred_edge_paths"].append(
                            edge_node_base_url + "/" + edge_node_id
                        )
                ha_vip_configs = locale_service.get("ha_vip_configs")
                if ha_vip_configs:
                    resource_params["ha_vip_configs"] = []
                    for ha_vip_config in ha_vip_configs:
                        external_interface_info = ha_vip_config.get("external_interface_info")
                        external_interface_paths = []
                        for external_interface in external_interface_info:
                            interface_base_url = NSXTTier0.NSXTTier0LocaleService.NSXTTier0Interface.get_resource_base_url(
                                self.get_parent_info()
                            )
                            external_interface_id = external_interface.get("external_interface_id")
                            external_interface_paths.append(
                                interface_base_url + "/" + external_interface_id
                            )
                        ha_vip_config["external_interface_paths"] = external_interface_paths
                        resource_params["ha_vip_configs"].append(ha_vip_config)
                if not "id" in locale_service:
                    resource_params["id"] = resource_params["display_name"]
                self.multi_resource_params.append(resource_params)

        def update_parent_info(self, parent_info):
            parent_info["locale_services_id"] = self.resource_params["id"]
            parent_info["ls_display_name"] = self.resource_params["display_name"]

        class NSXTTier0Interface(NSXTPolicyBaseResource):
            def get_spec_identifier(self):
                return NSXTTier0.NSXTTier0LocaleService.NSXTTier0Interface.get_spec_identifier()

            @classmethod
            def get_spec_identifier(cls):
                return "interfaces"

            @staticmethod
            def get_resource_base_url(parent_info):
                tier0_id = parent_info.get("tier0_id", "default")
                locale_service_id = parent_info.get("locale_services_id", "default")
                return TIER_0_LS_INTERFACE_URL.format(tier0_id, locale_service_id)

            def update_resource_params(self, **kwargs):
                self.multi_resource_params = []
                fields = {
                    "access_vlan_id",
                    "description",
                    "dhcp_relay_id",
                    "display_name",
                    "id",
                    "igmp_local_join_groups",
                    "ls_id",
                    "mtu",
                    "multicast",
                    "ospf",
                    "proxy_arp_filters",
                    "resource_type",
                    "subnets",
                    "tags",
                    "state",
                    "type",
                    "urpf_mode",
                }
                locale_services = kwargs.get("locale_services")
                ls_display_name = self._parent_info.get("ls_display_name")

                locale_service = next(
                    (ls for ls in locale_services if ls.get("display_name") == ls_display_name),
                    {},
                )
                if locale_service:
                    interfaces = locale_service.get("interfaces") or {}
                    # This block can be refactored
                    for interface in interfaces:
                        resource_params = {}
                        for field in fields:
                            val = interface.get(field)
                            if val:
                                resource_params[field] = val
                        ipv6_profile_paths = []
                        ipv6_ndra_profile_id = interface.get("ipv6_ndra_profile_id")
                        if ipv6_ndra_profile_id:
                            ipv6_profile_paths.append(
                                IPV6_NDRA_PROFILE_URL + "/" + interface.get("ipv6_ndra_profile_id")
                            )
                            resource_params["ipv6_profile_paths"] = ipv6_profile_paths
                        # segment_id is a required attr
                        segment_id = interface.get("segment_id")
                        if not segment_id:
                            raise SaltInvocationError(
                                {
                                    "resourceType": "Tier0Interface",
                                    "error": "required attribute segment_id not found",
                                }
                            )
                        resource_params["segment_path"] = SEGMENT_URL + "/" + segment_id
                        # edge_node_info is a required attr
                        edge_node_info = interface.get("edge_node_info")
                        if not edge_node_info:
                            raise SaltInvocationError(
                                {
                                    "resourceType": "Tier0Interface",
                                    "error": "required attribute edge_node_info not found",
                                }
                            )
                        edge_node_info = interface.get("edge_node_info")
                        edge_node_id = edge_node_info.get("edge_node_id")
                        edge_node_base_url = EDGE_NODE_URL.format(
                            edge_node_info["site_id"],
                            edge_node_info["enforcementpoint_id"],
                            edge_node_info["edge_cluster_id"],
                        )
                        resource_params["edge_path"] = edge_node_base_url + "/" + edge_node_id
                        resource_params["resource_type"] = "ServiceInterface"
                        if not resource_params.get("id"):
                            resource_params["id"] = resource_params["display_name"]
                        self.multi_resource_params.append(resource_params)

        class NSXTTier0LocaleServiceBGP(NSXTPolicyBaseResource):
            def __init__(self):
                self.id = "bgp"
                super().__init__()

            def skip_delete(self):
                return True

            def get_spec_identifier(self):
                return (
                    NSXTTier0.NSXTTier0LocaleService.NSXTTier0LocaleServiceBGP.get_spec_identifier()
                )

            @classmethod
            def get_spec_identifier(cls):
                return "BGP"

            @classmethod
            def is_object_deletable(cls):
                return False

            def update_resource_params(self, **kwargs):
                self.multi_resource_params = []
                resource_params = {}
                fields = {
                    "description",
                    "display_name",
                    "ecmp",
                    "enabled",
                    "graceful_restart_config",
                    "id",
                    "inter_sr_ibgp",
                    "local_as_num",
                    "multipath_relax",
                    "state",
                    "route_aggregations",
                    "tags",
                }
                locale_services = kwargs.get("locale_services") or {}
                ls_display_name = self._parent_info.get("ls_display_name")

                locale_service = next(
                    (ls for ls in locale_services if ls.get("display_name") == ls_display_name),
                    {},
                )
                bgp = locale_service.get("bgp") or {}
                if bgp:
                    # This pattern of code is identical in pattern and can be refactored.
                    for field in fields:
                        val = bgp.get(field)
                        if val:
                            resource_params[field] = val
                    resource_params["resource_type"] = "BgpRoutingConfig"
                    resource_params["id"] = "bgp"
                    self.multi_resource_params.append(resource_params)

            @staticmethod
            def get_resource_base_url(parent_info):
                tier0_id = parent_info.get("tier0_id", "default")
                locale_service_id = parent_info.get("locale_services_id", "default")
                return (TIER_0_LOCALE_SERVICE_URL + "/{}").format(tier0_id, locale_service_id)

            class NSXTTier0LocaleServiceBGPNeighbor(NSXTPolicyBaseResource):
                def get_spec_identifier(self):
                    return (
                        NSXTTier0.NSXTTier0LocaleService.NSXTTier0LocaleServiceBGP.get_spec_identifier()
                    )

                @classmethod
                def get_spec_identifier(cls):
                    return "neighbors"

                def update_resource_params(self, **kwargs):
                    self.multi_resource_params = []
                    fields = {
                        "tags",
                        "source_addresses",
                        "route_filtering",
                        "remote_as_num",
                        "password",
                        "neighbor_address",
                        "maximum_hop_limit",
                        "keep_alive_time",
                        "id",
                        "hold_down_time",
                        "graceful_restart_mode",
                        "state",
                        "display_name",
                        "description",
                        "bfd",
                        "allow_as_in",
                    }
                    locale_services = kwargs.get("locale_services") or {}
                    ls_display_name = self._parent_info.get("ls_display_name")
                    locale_service = next(
                        (ls for ls in locale_services if ls.get("display_name") == ls_display_name),
                        {},
                    )
                    if locale_service:
                        neighbors = locale_service.get("bgp").get("neighbors") or {}
                        for neighbor in neighbors:
                            resource_params = {}
                            # This pattern of code is identical in pattern and can be refactored.
                            for field in fields:
                                val = neighbor.get(field)
                                if val:
                                    resource_params[field] = val
                            resource_params["resource_type"] = "BgpNeighborConfig"
                            if not resource_params.get("id"):
                                resource_params["id"] = resource_params["display_name"]
                            self.multi_resource_params.append(resource_params)

                @staticmethod
                def get_resource_base_url(parent_info):
                    tier0_id = parent_info.get("tier0_id", "default")
                    locale_service_id = parent_info.get("locale_services_id", "default")
                    return TIER_0_BGP_NEIGHBOR_URL.format(tier0_id, locale_service_id)


def get_by_display_name(
    hostname, username, password, display_name, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Gets Tier 0 Gateway present in the NSX-T Manager with given name.

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_policy_tier0.get_by_display_name hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        The name of Tier 0 Gateway to fetch

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """
    nsxt_tier0 = NSXTTier0()
    url = (NSXTPolicyBaseResource.get_nsxt_base_url() + nsxt_tier0.get_resource_base_url()).format(
        hostname
    )
    return nsxt_tier0.get_by_display_name(
        url,
        username,
        password,
        display_name,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )


def get(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    cursor=None,
    included_fields=None,
    page_size=None,
    sort_ascending=None,
    sort_by=None,
):
    """
    Lists NSXT Tier 0 Gateways present in the NSX-T Manager

    CLI Example:
    .. code-block:: bash

        salt vm_minion nsxt_policy_tier0.get hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    include_mark_for_delete_objects
        (Optional) Include objects that are marked for deletion in results. If true, resources that are marked for
        deletion will be included in the results. By default, these resources are not included.

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    page_size
        (Optional) Maximum number of results to return in this page

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order
    """
    nsxt_tier0 = NSXTTier0()
    url = (NSXTPolicyBaseResource.get_nsxt_base_url() + nsxt_tier0.get_resource_base_url()).format(
        hostname
    )
    return nsxt_tier0.get(
        url,
        username,
        password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        cursor=cursor,
        included_fields=included_fields,
        page_size=page_size,
        sort_ascending=sort_ascending,
        sort_by=sort_by,
    )


def create_or_update(
    hostname,
    username,
    password,
    cert=None,
    cert_common_name=None,
    verify_ssl=True,
    arp_limit=None,
    bfd_peers=None,
    display_name=None,
    description=None,
    default_rule_logging=None,
    dhcp_config_id=None,
    disable_firewall=None,
    failover_mode=None,
    force_whitelisting=None,
    ha_mode=None,
    id=None,
    internal_transit_subnets=None,
    intersite_config=None,
    ipv6_ndra_profile_id=None,
    ipv6_dad_profile_id=None,
    locale_services=None,
    rd_admin_field=None,
    static_routes=None,
    tags=None,
    transit_subnets=None,
    vrf_config=None,
):
    """
    Creates a Tier 0 Gateway and its sub-resources with given specifications

    CLI Example:
    .. code-block:: bash

        salt vm_minion nsxt_policy_tier0.create hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    display_name:
        description:
            - Display name.
            - If resource ID is not specified, display_name will be used as ID.

        required: false
        type: str

    tags:
        description: Opaque identifiers meaningful to the API user.

        type: dict

        suboptions:

            scope:
                description: Tag scope.
                required: true
                type: str

            tag:
                description: Tag value.
                required: true
                type: str

    id:

        description: Tier-0 ID

        required: false

        type: str

    description:

        description: Tier-0 description

        type: str

    state:

        description: present or absent keyword is used as an indetifier, default value is present.
                    If a user has provided absent that resource/sub-resource will be deleted

    default_rule_logging:

        description: Enable logging for whitelisted rule.
                     Indicates if logging should be enabled for the default
                     whitelisting rule.

        default: false

    ha_mode:

        description: High-availability Mode for Tier-0

        choices:
            - 'ACTIVE_STANDBY'
            - 'ACTIVE_ACTIVE'

        default: 'ACTIVE_ACTIVE'

        type: str

    disable_firewall:

        description: Disable or enable gateway fiewall.

        default: False

        type: bool

    failover_mode:

        description: Determines the behavior when a Tier-0 instance in
                     ACTIVE-STANDBY high-availability mode restarts
                     after a failure. If set to PREEMPTIVE, the preferred node
                     will take over, even if it causes
                     another failure. If set to NON_PREEMPTIVE, then
                     the instance that restarted will remain secondary.
                     This property must not be populated unless the
                     ha_mode property is set to ACTIVE_STANDBY.

        choices:

            - 'NON_PREEMPTIVE'
            - 'PREEMPTIVE'

        default: 'NON_PREEMPTIVE'

        type: str

    force_whitelisting:

        description: Flag to add whitelisting FW rule during
                     realization.

        default: False

        type: bool

    internal_transit_subnets:

        description: Internal transit subnets in CIDR format.
                     Specify subnets that are used to assign addresses
                     to logical links connecting service routers and
                     distributed routers. Only IPv4 addresses are
                     supported. When not specified, subnet 169.254.0.0/
                     24 is assigned by default in ACTIVE_ACTIVE HA mode
                     or 169.254.0.0/28 in ACTIVE_STANDBY mode.

        default: False

        type: list

    intersite_config:

        description: Inter site routing configuration when the gateway is
                     streched.

        type: dict

        suboptions:

            fallback_sites:
                description: Fallback site to be used as new primary
                             site on current primary site failure.
                             Disaster recovery must be initiated via
                             API/UI. Fallback site configuration is
                             supported only for T0 gateway. T1 gateway
                             will follow T0 gateway's primary site
                             during disaster recovery

                type: list

            intersite_transit_subnet:
                description:
                    - Transit subnet in CIDR format
                    - IPv4 subnet for inter-site transit segment
                      connecting service routers across sites for
                      stretched gateway. For IPv6 link local subnet is
                      auto configured

                type: str
                default: "169.254.32.0/20"

            last_admin_active_epoch:
                description:
                    - Epoch of last time admin changing active
                      LocaleServices
                    - Epoch(in seconds) is auto updated based on
                      system current timestamp when primary locale
                      service is updated. It is used for resolving
                      conflict during site failover. If system clock
                      not in sync then User can optionally override
                      this. New value must be higher than the current
                      value.

                type: int

            primary_site_path:
                description:
                    - Primary egress site for gateway.
                    - Primary egress site for gateway. T0/T1 gateway in
                      Active/Standby mode supports stateful services on primary
                      site. In this mode primary site must be set if gateway is
                      stretched to more than one site. For T0 gateway in
                      Active/Active primary site is optional field. If set then
                      secondary site prefers routes learned from primary over
                      locally learned routes. This field is not applicable for
                      T1 gateway with no services

                type: str

    ipv6_ndra_profile_id:

        description: IPv6 NDRA profile configuration on Tier0.
                     Either or both NDRA and/or DAD profiles can be
                     configured. Related attribute ipv6_dad_profile_id.

        type: str

    ipv6_dad_profile_id:

        description: IPv6 DRA profile configuration on Tier0.
                     Either or both NDRA and/or DAD profiles can be
                     configured. Related attribute ipv6_ndra_profile_id.

    rd_admin_field:

        description:
            - Route distinguisher administrator address
            - If you are using EVPN service, then route distinguisher
              administrator address should be defined if you need auto
              generation of route distinguisher on your VRF configuration

        type: str

    transit_subnets:

        description: Transit subnets in CIDR format.
                     Specify transit subnets that are used to assign
                     addresses to logical links connecting tier-0 and
                     tier-1s. Both IPv4 and IPv6 addresses are
                     supported.
                     When not specified, subnet 100.64.0.0/16 is
                     configured by default.

        type: list

    dhcp_config_id:

        description: DHCP configuration for Segments connected to
                     Tier-0. DHCP service is configured in relay mode.

        type: str

    vrf_config:

        type: dict
        description: VRF config, required for VRF Tier0

        suboptions:

            evpn_transit_vni:

                description:
                    - L3 VNI associated with the VRF for overlay traffic.
                    - VNI must be unique and belong to configured VNI pool.

                type: int

            route_distinguisher:
                description: Route distinguisher. 'ASN:<>' or 'IPAddress:<>'.

                type: str

            route_targets:
                description: Route targets
                type: list
                element: dict
                suboptions:

                    export_route_targets:
                        description: Export route targets. 'ASN:' or
                                     'IPAddress:<>'

                        type: list

                        element: str

                    import_route_targets:
                        description: Import route targets. 'ASN:' or
                                     'IPAddress:<>'

                        type: list

                        element: str

            tier0_id:

                description: Default tier0 id. Cannot be modified after
                             realization. Either this or tier0_id must
                             be specified

                type: str

    static_routes:

        type: list
        element: dict
        description: This is a list of Static Routes that need to be created, updated, or deleted
        suboptions:

            id:
                description: Tier-0 Static Route ID.
                required: false
                type: str

            display_name:

                description:
                    - Tier-0 Static Route display name.
                    - Either this or id must be specified. If both are
                      specified, id takes precedence.

                required: false

                type: str

            description:

                description:
                    - Tier-0 Static Route description.

                type: str

            state:
                description: present or absent keyword is used as an indetifier, default value is present.
                             If a user has provided absent that resource/sub-resource will be deleted

            network:

                description: Network address in CIDR format
                required: true
                type: str

            next_hops:

                description: Next hop routes for network
                type: list
                elements: dict
                suboptions:

                    admin_distance:
                        description: Cost associated with next hop route
                        type: int
                        default: 1

                    ip_address:
                        description: Next hop gateway IP address
                        type: str

                    scope:
                        description:
                            - Interface path associated with current route
                            - For example, specify a policy path referencing the
                              IPSec VPN Session

                        type: list

            tags:

                description: Opaque identifiers meaningful to the API user

                type: dict

                suboptions:

                    scope:
                        description: Tag scope.
                        required: true
                        type: str

                    tag:
                        description: Tag value.
                        required: true
                        type: str

    bfd_peers:

        type: list
        element: dict
        description: This is a list of BFD Peers that need to be created, updated, or deleted
        suboptions:

            id:
                description: Tier-0 BFD Peer ID.
                required: false
                type: str

            display_name:
                description:
                    - Tier-0 BFD Peer display name.
                    - Either this or id must be specified. If both are
                      specified, id takes precedence.

                required: false
                type: str

            description:
                description:
                    - Tier-0 BFD Peer description. config

                type: str

            state:
                description: present or absent keyword is used as an indetifier, default value is present.
                             If a user has provided absent that resource/sub-resource will be deleted

            bfd_profile_id:

                description:
                    - The associated BFD Profile ID
                    - Either this or bfd_profile_display_name must be specified
                    - BFD Profile is not supported for IPv6 networks.

                type: str

            enabled:
                description: Flag to enable BFD peer.
                type: boolean

            peer_address:
                description: IP Address of static route next hop peer. Only IPv4 addresses are supported
                             Only a single BFD config per peer address is allowed.

                type: str

            source_addresses:
                description: List of source IP addresses. Array of Tier0 external interface IP addresses. BFD peering
                             is established from all these source addresses to the neighbor specified in peer_address.
                             Only IPv4 addresses are supported.(Minimum-0, Maximum-8 values are allowed)

                type: list
                elements: IPv4 addresse strings

            scope:
                description: Array of policy paths of locale services. Represents the array of policy paths of
                             locale services where this BFD peer should get relalized on. The locale service service
                             and this BFD peer must belong to the same router. Default scope is empty.

                type: list
                elements: policy path string of locale services

            tags:
                description: Opaque identifiers meaningful to the API user

                type: dict

                suboptions:

                    scope:
                        description: Tag scope.
                        required: true
                        type: str
                    tag:
                        description: Tag value.
                        required: true
                        type: str

    locale_services:

        type: list
        element: dict
        description: This is a list of Locale Services that need to be created,updated, or deleted

        suboptions:

            id:
                description: Tier-0 Locale Service ID.
                required: false
                type: str

            display_name:
                description:
                    - Tier-0 Locale Service display name.
                    - Either this or id must be specified. If both are
                      specified, id takes precedence

                required: false

                type: str

            description:
                description:
                    - Tier-0 Locale Service  description.

                type: str

            state:
                description: present or absent keyword is used as an indetifier, default value is present.
                             If a user has provided absent that resource/sub-resource will be deleted

            tags:
                description: Opaque identifiers meaningful to the API user
                type: dict
                suboptions:

                    scope:
                        description: Tag scope.
                        required: true
                        type: str

                    tag:
                        description: Tag value.
                        required: true
                        type: str

            edge_cluster_info:
                description: Used to create path to edge cluster. Auto-assigned
                            if associated enforcement-point has only one edge
                            cluster.

                type: dict

                suboptions:

                    site_id:

                        description: site_id where edge cluster is located
                        default: default
                        type: str

                    enforcementpoint_id:
                        description: enforcementpoint_id where edge cluster is located
                        default: default
                        type: str

                    edge_cluster_id:
                        description: ID of the edge cluster
                        type: str

            preferred_edge_nodes_info:

                description: Used to create paths to edge nodes. Specified edge
                            is used as preferred edge cluster member when
                            failover mode is set to PREEMPTIVE, not
                            applicable otherwise.

                type: list
                suboptions:

                    site_id:
                        description: site_id where edge node is located
                        default: default
                        type: str

                    enforcementpoint_id:
                        description: enforcementpoint_id where edge node is located
                        default: default
                        type: str

                    edge_cluster_id:
                        description: edge_cluster_id where edge node is located
                        type: str

                    edge_node_id:
                        description: ID of the edge node
                        type: str

            route_redistribution_config:

                description: Configure all route redistribution properties like
                             enable/disable redistributon, redistribution rule
                             and so on.

                type: dict

                suboptions:

                    bgp_enabled:
                        description: Flag to enable route redistribution.
                        type: bool
                        default: false

                    redistribution_rules:
                        description: List of redistribution rules.
                        type: list
                        elements: dict
                        suboptions:

                            name:
                                description: Rule name
                                type: str

                            route_map_path:

                                description: Route map to be associated with
                                             the redistribution rule

                                type: str

                            route_redistribution_types:

                                description: Tier-0 route redistribution types
                                choices:

                                    - TIER0_STATIC - Redistribute user added
                                      static routes.
                                    - TIER0_CONNECTED - Redistribute all
                                      subnets configured on Interfaces and
                                      routes related to TIER0_ROUTER_LINK,
                                      TIER0_SEGMENT, TIER0_DNS_FORWARDER_IP,
                                      TIER0_IPSEC_LOCAL_IP, TIER0_NAT types.
                                    - TIER1_STATIC - Redistribute all subnets
                                      and static routes advertised by Tier-1s.
                                    - TIER0_EXTERNAL_INTERFACE - Redistribute
                                      external interface subnets on Tier-0.
                                    - TIER0_LOOPBACK_INTERFACE - Redistribute
                                      loopback interface subnets on Tier-0.
                                    - TIER0_SEGMENT - Redistribute subnets
                                      configured on Segments connected to
                                      Tier-0.
                                    - TIER0_ROUTER_LINK - Redistribute router
                                      link port subnets on Tier-0.
                                    - TIER0_SERVICE_INTERFACE - Redistribute
                                      Tier0 service interface subnets.
                                    - TIER0_DNS_FORWARDER_IP - Redistribute DNS
                                      forwarder subnets.
                                    - TIER0_IPSEC_LOCAL_IP - Redistribute IPSec
                                      subnets.
                                    - TIER0_NAT - Redistribute NAT IPs owned by
                                      Tier-0.
                                    - TIER0_EVPN_TEP_IP - Redistribute EVPN
                                      local endpoint subnets on Tier-0.
                                    - TIER1_NAT - Redistribute NAT IPs
                                      advertised by Tier-1 instances.
                                    - TIER1_LB_VIP - Redistribute LB VIP IPs
                                      advertised by Tier-1 instances.
                                    - TIER1_LB_SNAT - Redistribute LB SNAT IPs
                                      advertised by Tier-1 instances.
                                    - TIER1_DNS_FORWARDER_IP - Redistribute DNS
                                      forwarder subnets on Tier-1 instances.
                                    - TIER1_CONNECTED - Redistribute all
                                      subnets configured on Segments and
                                      Service Interfaces.
                                    - TIER1_SERVICE_INTERFACE - Redistribute
                                      Tier1 service interface subnets.
                                    - TIER1_SEGMENT - Redistribute subnets
                                      configured on Segments connected to
                                      Tier1.
                                    - TIER1_IPSEC_LOCAL_ENDPOINT - Redistribute
                                      IPSec VPN local-endpoint subnets
                                      advertised by TIER1.

                                type: list

            ha_vip_configs:

                type: list

                elements: dict

                description:

                    - Array of HA VIP Config.
                    - This configuration can be defined only for Active-Standby
                      Tier0 gateway to provide redundancy. For mulitple
                      external interfaces, multiple HA VIP configs must be
                      defined and each config will pair exactly two external
                      interfaces. The VIP will move and will always be owned by
                      the Active node. When this property is configured,
                      configuration of dynamic-routing is not allowed.

                suboptions:

                    enabled:
                        description: Flag to enable this HA VIP config.
                        default: true
                        type: bool

                    external_interface_paths:
                        description:
                            - Policy paths to Tier0 external interfaces for
                              providing redundancy
                            - Policy paths to Tier0 external interfaces which
                              are to be paired to provide redundancy. Floating
                              IP will be owned by one of these interfaces
                              depending upon which edge node is Active.

                        type: list

                    vip_subnets:

                        description:
                            - VIP floating IP address subnets
                            - Array of IP address subnets which will be used as
                              floating IP addresses.

                        type: list

                        suboptions:

                            ip_addresses:
                                description: IP addresses assigned to interface
                                type: list
                                required: true

                            prefix_len:
                                description: Subnet prefix length
                                type: int
                                required: true

            bgp:

                description: Specify the BGP spec in this section

                type: dict

                state:

                    description: present or absent keyword is used as an indetifier, default value is present,
                                 If a user has provided absent that resource/sub-resource will be deleted.

                suboptions:

                    ecmp:
                        description: Flag to enable ECMP.
                        type: bool
                        required: False
                        default: True

                    enabled:
                        description: Flag to enable BGP configuration.
                                     Disabling will stop feature and BGP
                                     peering.

                        type: bool

                        default: True

                    graceful_restart_config:
                        description: Configuration field to hold BGP Restart
                                     mode and timer.

                        type: dict
                        required: False
                        suboptions:

                            mode:

                                description:
                                    - BGP Graceful Restart Configuration Mode
                                    - If mode is DISABLE, then graceful restart
                                      and helper modes are disabled.
                                    - If mode is GR_AND_HELPER, then both
                                      graceful restart and helper modes are
                                      enabled.
                                    - If mode is HELPER_ONLY, then helper mode
                                      is enabled. HELPER_ONLY mode is the
                                      ability for a BGP speaker to indicate its
                                      ability to preserve forwarding state
                                      during BGP restart.
                                    - GRACEFUL_RESTART mode is the ability of a
                                      BGP speaker to advertise its restart to
                                      its peers.

                                type: str

                                required: False

                                default: 'HELPER_ONLY'

                                choices:
                                    - DISABLE
                                    - GR_AND_HELPER
                                    - HELPER_ONLY

                            timer:
                                description: BGP Graceful Restart Timer
                                type: dict
                                required: False
                                suboptions:

                                    restart_timer:

                                        description:
                                            - BGP Graceful Restart Timer
                                            - Maximum time taken (in seconds)
                                              for a BGP session to be
                                              established after a restart. This
                                              can be used to speed up routing
                                              convergence by its peer in case
                                              the BGP speaker does not come
                                              back up after a restart. If the
                                              session is not re-established
                                              within this timer, the receiving
                                              speaker will delete all the stale
                                              routes from that peer. Min 1 and
                                              Max 3600

                                        type: int

                                        default: 180

                                    stale_route_timer:

                                        description:

                                            - BGP Stale Route Timer
                                            - Maximum time (in seconds) before
                                              stale routes are removed from the
                                              RIB (Routing Information Base)
                                              when BGP restarts. Min 1 and Max
                                              3600

                                        type: int

                                        default: 600

                    inter_sr_ibgp:

                        description: Flag to enable inter SR IBGP
                                     configuration. When not specified, inter
                                     SR IBGP is automatically enabled if Tier-0
                                     is created in ACTIVE_ACTIVE ha_mode.

                        type: bool
                        required: False

                    local_as_num:

                        description:

                            - BGP AS number in ASPLAIN/ASDOT Format.
                            - Specify BGP AS number for Tier-0 to advertize to
                              BGP peers. AS number can be specified in ASPLAIN
                              (e.g., "65546") or ASDOT (e.g., "1.10") format.
                              Empty string disables BGP feature.

                        type: str
                        required: True

                    multipath_relax:
                        description: Flag to enable BGP multipath relax option.
                        type: bool
                        default: True

                    route_aggregations:
                        description: List of routes to be aggregated
                        type: dict
                        required: False
                        suboptions:

                            prefix:

                                description: CIDR of aggregate address
                                type: str
                                required: True

                            summary_only:
                                description:
                                    - Send only summarized route.
                                    - Summarization reduces number of routes
                                      advertised by representing multiple
                                      related routes with prefix property

                                type: bool
                                default: True

                    neighbors:

                        description: Specify the BGP neighbors in this section
                                     that need to be created, updated, or
                                     deleted

                        type: list
                        element: dict
                        state:

                            description: present or absent keyword is used as an indetifier, default value is present.
                                         If a user has provided absent that resource/sub-resource will be deleted

                        suboptions:

                            allow_as_in:
                                description: Flag to enable allowas_in option
                                             for BGP neighbor.

                                type: bool
                                default: False

                            bfd:
                                description:
                                    - BFD configuration for failure detection
                                    - BFD is enabled with default values when
                                      not configured

                                type: dict
                                required: False

                                suboptions:

                                    enabled:
                                        description: Flag to enable BFD
                                                     cofiguration.

                                        type: bool
                                        required: False

                                    interval:
                                        description: Time interval between
                                                     heartbeat packets in
                                                     milliseconds. Min 300 and
                                                     Max 60000.

                                        type: int
                                        default: 1000

                                    multiple:
                                        description:
                                            - Declare dead multiple.
                                            - Number of times heartbeat packet
                                              is missed before BFD declares the
                                              neighbor is down.
                                              Min 2 and Max 16.

                                        type: int
                                        default: 3

                            graceful_restart_mode:

                                description:
                                    - BGP Graceful Restart Configuration Mode
                                    - If mode is DISABLE, then graceful restart
                                      and helper modes are disabled.
                                    - If mode is GR_AND_HELPER, then both
                                      graceful restart and helper modes are
                                      enabled.
                                    - If mode is HELPER_ONLY, then helper mode
                                      is enabled. HELPER_ONLY mode is the
                                      ability for a BGP speaker to indicate its
                                      ability to preserve forwarding state
                                      during BGP restart.
                                    - GRACEFUL_RESTART mode is the ability of a
                                      BGP speaker to advertise its restart to
                                      its peers.

                                type: str

                                choices:
                                    - DISABLE
                                    - GR_AND_HELPER
                                    - HELPER_ONLY

                            hold_down_time:

                                description: Wait time in seconds before
                                             declaring peer dead. Min 1 and Max
                                             65535.

                                type: int
                                default: 180

                            keep_alive_time:

                                description: Interval between keep alive
                                             messages sent to peer. Min 1 and
                                             Max 65535.

                                type: int
                                default: 60

                            maximum_hop_limit:

                                description: Maximum number of hops allowed to
                                             reach BGP neighbor. Min 1 and Max
                                             255.

                                type: int
                                default: 1

                            address:

                                description: Neighbor IP Address

                                type: str
                                required: True

                            password:

                                description: Password for BGP Neighbor
                                             authentication. Empty string ("")
                                             clears existing password.

                                type: str
                                required: False

                            remote_as_num:

                                description: 4 Byte ASN of the neighbor in
                                             ASPLAIN Format.

                                type: str
                                required: True

                            route_filtering:

                                description: Enable address families and route
                                             filtering in each direction.

                                type: list
                                elements: dict
                                required: False
                                suboptions:

                                    address_family:

                                        type: str
                                        required: False
                                        choices:

                                            - 'IPV4'
                                            - 'IPV6'
                                            - 'VPN'

                                    enabled:

                                        description: Flag to enable address
                                                     family.

                                        type: bool
                                        default: True

                                    in_route_filters:

                                        description:

                                            - Prefix-list or route map path for
                                              IN direction
                                            - Specify path of prefix-list or
                                              route map to filter routes for IN
                                              direction.

                                        type: list
                                        required: False

                                    out_route_filters:

                                        description:

                                            - Prefix-list or route map path for
                                              OUT direction
                                            - Specify path of prefix-list or
                                              route map to filter routes
                                              for OUT direction. When not
                                              specified, a built-in
                                              prefix-list named
                                              'prefixlist-out-default' is
                                              automatically applied.

                                        type: list
                                        required: False

                            source_addresses:

                                description:
                                    - Source IP Addresses for BGP peering
                                    - Source addresses should belong to Tier0
                                      external or loopback interface IP
                                      Addresses. BGP peering is formed from all
                                      these addresses. This property is
                                      mandatory when maximum_hop_limit is
                                      greater than 1.

                                type: list
                                required: False

            interfaces:

                type: list

                element: dict

                description: Specify the interfaces associated with the Gateway
                             in this section that need to be created, updated,
                             or deleted

                state:

                    description: present or absent keyword is used as an indetifier, default value is present.
                                 If a user has provided absent that resource/sub-resource will be deleted

                suboptions:

                    id:
                        description: Tier-0 Interface ID
                        type: str

                    display_name:
                        description:
                            - Tier-0 Interface display name
                            - Either this or id must be specified. If both are
                              specified, id takes precedence.

                        required: false
                        type: str

                    description:

                        description: Tier-0 Interface  description
                        type: str

                    state:

                        description:
                            - State can be either 'present' or 'absent'.
                              'present' is used to create or update resource.
                              'absent' is used to delete resource.
                            - Required if I(segp_id != null)

                        choices:

                            - present
                            - absent

                    tags:
                        description: Opaque identifiers meaningful to the API
                                     user.

                        type: dict

                        suboptions:

                            scope:
                                description: Tag scope.
                                required: true
                                type: str

                            tag:
                                description: Tag value.
                                required: true
                                type: str

                    access_vlan_id:

                        description: Vlan id
                        type: int

                    ipv6_ndra_profile_display_name:

                        description: Same as ipv6_ndra_profile_id. Either one
                                     should be specified.

                        type: str

                    ipv6_ndra_profile_id:

                        description: Configuration IPv6 NDRA profile. Only one
                                     NDRA profile can be configured.

                        type: str

                    mtu:
                        description:
                            - MTU size
                            - Maximum transmission unit (MTU) specifies the
                              size of the largest packet that a network
                              protocol can transmit.

                        type: int

                    multicast:
                        description: Multicast PIM configuration

                        type: dict
                        suboptions:

                            enabled:

                                description: enable/disable PIM configuration
                                type: bool
                                default: False
                    urpf_mode:
                        description: Unicast Reverse Path Forwarding mode

                        type: str
                        choices:

                            - NONE
                            - STRICT

                    segment_id:
                        description: Specify Segment to which this interface is
                                     connected to. Required if id is specified.

                        type: str

                    segment_display_name:
                        description:
                            - Same as segment_id
                            - Either this or segment_id must be specified. If
                              both are specified, segment_id takes precedence.

                        type: str

                    type:

                        description: Interface type

                        choices:

                            - "EXTERNAL"
                            - "LOOPBACK"
                            - "SERVICE"

                        type: str

                    edge_node_info:

                        description:
                            - Info to create policy path to edge node to
                              handle externalconnectivity.
                            - Required if interface type is EXTERNAL and
                              I(id != null)

                        type: dict
                        suboptions:

                            site_id:
                                description: site_id where edge node is located

                                default: default
                                type: str

                            enforcementpoint_id:
                                description: enforcementpoint_id where edge
                                             node is located.

                                default: default
                                type: str

                            edge_cluster_id:
                                description: edge_cluster_id where edge node is
                                             located

                                type: str

                            edge_node_id:
                                description: ID of the edge node

                                type: str

                    subnets:
                        description:
                            - IP address and subnet specification for interface
                            - Specify IP address and network prefix for
                              interface.
                            - Required if I(id != null).

                        required: False
                        type: list
    """
    execution_logs = []
    try:
        nsxt_tier0 = NSXTTier0()
        nsxt_tier0.create_or_update(
            hostname=hostname,
            username=username,
            password=password,
            execution_logs=execution_logs,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
            arp_limit=arp_limit,
            bfd_peers=bfd_peers,
            description=description,
            display_name=display_name,
            default_rule_logging=default_rule_logging,
            dhcp_config_id=dhcp_config_id,
            disable_firewall=disable_firewall,
            failover_mode=failover_mode,
            force_whitelisting=force_whitelisting,
            ha_mode=ha_mode,
            id=id,
            internal_transit_subnets=internal_transit_subnets,
            intersite_config=intersite_config,
            ipv6_ndra_profile_id=ipv6_ndra_profile_id,
            ipv6_dad_profile_id=ipv6_dad_profile_id,
            locale_services=locale_services,
            rd_admin_field=rd_admin_field,
            static_routes=static_routes,
            tags=tags,
            transit_subnets=transit_subnets,
            vrf_config=vrf_config,
        )
        return execution_logs
    except SaltInvocationError as e:
        execution_logs.append({"error": str(e)})
        return execution_logs


def delete(
    hostname, username, password, tier0_id, cert=None, cert_common_name=None, verify_ssl=True
):
    """
    Deletes a Tier 0 gateway and it sub-resources

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    tier0_id
        id of the tier 0 to be deleted

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """
    execution_logs = []
    try:
        nsxt_tier0 = NSXTTier0()
        nsxt_tier0.delete(
            hostname,
            username,
            password,
            tier0_id,
            cert,
            cert_common_name,
            verify_ssl,
            execution_logs,
        )
        return execution_logs
    except SaltInvocationError as e:
        execution_logs.append({"error": str(e)})
        return execution_logs


def get_hierarchy(
    hostname, username, password, tier0_id, cert=None, cert_common_name=None, verify_ssl=True
):
    """
    Returns entire hierarchy of Tier 0 gateway and its sub-resources

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    tier0_id
        id of the tier 0 gateway

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """
    result = {}
    try:
        nsxt_tier0 = NSXTTier0()
        nsxt_tier0.get_hierarchy(
            hostname, username, password, tier0_id, cert, cert_common_name, verify_ssl, result
        )
        log.info("Hierarchy result for tier 0 gateway: {}".format(result))
        return result
    except SaltInvocationError as e:
        return {"error": str(e)}



"""
Execution module for NSX-T compute manager registration and de-registration
"""
import json
import logging

from saltext.vmware.utils import common
from saltext.vmware.utils import nsxt_request

log = logging.getLogger(__name__)

__virtualname__ = "nsxt_compute_manager"

BASE_URL = "https://{management_host}/api/v1/fabric/compute-managers"


def __virtual__():
    return __virtualname__


def get(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    cursor=None,
    included_fields=None,
    origin_type=None,
    page_size=None,
    server=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Lists compute managers registered to NSX Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_compute_manager.get hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    origin_type
        (Optional) Compute manager type like vCenter

    page_size
        (Optional) Maximum number of results to return in this page

    server
        (Optional) IP address or hostname of compute manager

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order

    """

    url = BASE_URL.format(management_host=hostname)
    log.info("Retrieving compute managers from NSX Manager {}".format(hostname))

    params = common._filter_kwargs(
        allowed_kwargs=(
            "server",
            "cursor",
            "included_fields",
            "origin_type",
            "page_size",
            "sort_by",
            "sort_ascending",
        ),
        server=server,
        cursor=cursor,
        included_fields=included_fields,
        origin_type=origin_type,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )

    return nsxt_request.call_api(
        method="get",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_display_name(
    hostname,
    username,
    password,
    display_name,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    cursor=None,
    included_fields=None,
    origin_type=None,
    page_size=None,
    server=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Lists compute managers registered to NSX Manager by given display_name

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_compute_manager.get_by_display_name hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        Display-name of the compute manager

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    origin_type
        (Optional) Compute manager type like vCenter

    page_size
        (Optional) Maximum number of results to return in this page

    server
        (Optional) IP address or hostname of compute manager

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order

    """

    log.info("Finding compute managers with display name: %s", display_name)

    compute_managers = common._read_paginated(
        func=get,
        display_name=display_name,
        cursor=cursor,
        included_fields=included_fields,
        origin_type=origin_type,
        page_size=page_size,
        server=server,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
        hostname=hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )

    if "error" in compute_managers:
        return compute_managers

    return {"results": compute_managers}


def register(
    hostname,
    username,
    password,
    compute_manager_server,
    credential,
    server_origin_type="vCenter",
    display_name=None,
    description=None,
    set_as_oidc_provider=None,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """
    Lists compute managers registered to NSX Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_compute_manager.register hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    compute_manager_server
        Compute manager server FQDN or IP

    credential
        An object which contains credential details to validate compute manager
        Sample usage in sls file:

        .. code::

            credential:
               credential_type: "UsernamePasswordLoginCredential"
               username: "user"
               password: "pass"
               thumbprint: "36:XX:XX:XX:XX:XX:XX66"

        credential_type
            Type of credential provided. For now only UsernamePasswordLoginCredential is supported.

        username
            Username of the compute manager

        password
            Password of the compute manager

        thumbprint
            Thumbprint of the compute manager

    server_origin_type
        (Optional) Server origin type of the compute manager. Default is vCenter

    display_name
        (Optional) Display name of the compute manager

    description
        (Optional) description for the compute manager

    set_as_oidc_provider
        (Optional) Specifies whether compute manager has been set as OIDC provider. Default is false
        If the compute manager is VC and need to set set as OIDC provider for NSX then this flag should be set as true.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """
    url = BASE_URL.format(management_host=hostname)
    log.info("Going to register new compute manager")

    data = common._filter_kwargs(
        allowed_kwargs=("description", "display_name", "set_as_oidc_provider"),
        default_dict={
            "server": compute_manager_server,
            "origin_type": server_origin_type,
            "credential": credential,
        },
        display_name=display_name,
        description=description,
        set_as_oidc_provider=set_as_oidc_provider,
    )

    return nsxt_request.call_api(
        method="post",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=data,
    )


def update(
    hostname,
    username,
    password,
    compute_manager_server,
    compute_manager_id,
    credential,
    compute_manager_revision,
    server_origin_type="vCenter",
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    display_name=None,
    description=None,
    set_as_oidc_provider=None,
):
    """
    Updates compute manager registered to NSX Manager

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_compute_manager.update hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    compute_manager_server
        Compute manager server FQDN or IP

    compute_manager_id
        Unique Id of the compute manager provided by NSX-T

    compute_manager_revision
        Latest value of _revision property for compute manager.

    credential
        An object which contains credential details to validate compute manager
        Sample usage in sls file:

        .. code::

            credential:
               credential_type: "UsernamePasswordLoginCredential"
               username: "user"
               password: "pass"
               thumbprint: "36:XX:XX:XX:XX:XX:XX66"

        credential_type
            Type of credential provided. For now only UsernamePasswordLoginCredential is supported.

        username
            Username of the compute manager

        password
            Password of the compute manager

        thumbprint
            Thumbprint of the compute manager

    server_origin_type
        (Optional) Server origin type of the compute manager. Default is vCenter

    display_name
        (Optional) Display name of the compute manager

    description
        (Optional) description for the compute manager

    set_as_oidc_provider
        (Optional) Specifies whether compute manager has been set as OIDC provider. Default is false
        If the compute manager is VC and need to set set as OIDC provider for NSX then this flag should be set as true.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """
    url = (BASE_URL + "/{compute_manager_id}").format(
        management_host=hostname, compute_manager_id=compute_manager_id
    )

    log.info("Going to update compute manager")
    data = {
        "server": compute_manager_server,
        "origin_type": server_origin_type,
        "credential": credential,
        "id": compute_manager_id,
        "_revision": compute_manager_revision,
    }
    optional_params = ("description", "display_name", "set_as_oidc_provider")

    data = common._filter_kwargs(
        allowed_kwargs=optional_params,
        default_dict=data,
        display_name=display_name,
        description=description,
        set_as_oidc_provider=set_as_oidc_provider,
    )

    return nsxt_request.call_api(
        method="put",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
        data=data,
    )


def remove(
    hostname,
    username,
    password,
    compute_manager_id,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
):
    """
    De-registers compute manager from NSX-T if exists

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_compute_manager.remove hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    compute_manager_id
        NSX-T unique id for the compute manager. Needed in case of updating compute manager details

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name verification.
        If the client certificate common name and hostname do not match (in case of self-signed certificates),
        specify the certificate common name as part of this parameter. This value is then used to compare against
        certificate common name.
    """

    url = (BASE_URL + "/{id}").format(management_host=hostname, id=compute_manager_id)
    log.info("Going to remove compute manager registration")
    result = nsxt_request.call_api(
        method="delete",
        url=url,
        username=username,
        password=password,
        cert_common_name=cert_common_name,
        verify_ssl=verify_ssl,
        cert=cert,
    )

    return result or {"message": "Removed compute manager successfully"}



import logging

import salt.exceptions
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.datacenter as utils_datacenter

# pylint: disable=no-name-in-module
try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

log = logging.getLogger(__name__)


def get_clusters(service_instance, datacenter_name=None, cluster_name=None):
    """
    Returns clusters in a vCenter.

    service_instance
        The Service Instance Object from which to obtain cluster.

    datacenter_name
        (Optional) Datacenter name to filter by.

    cluster_name
        (Optional) Exact cluster name to filter by. Requires datacenter_name.
    """
    if cluster_name and not datacenter_name:
        raise salt.exceptions.ArgumentValueError(
            "datacenter_name is required when looking up by cluster_name"
        )

    clusters = []
    for cluster in utils_common.get_mors_with_properties(
        service_instance, vim.ClusterComputeResource, property_list=["name"]
    ):
        if cluster_name and cluster_name != cluster["name"]:
            continue
        if (
            datacenter_name
            and datacenter_name
            != utils_common.get_parent_of_type(cluster["object"], vim.Datacenter).name
        ):
            continue
        clusters.append(cluster["object"])
    return clusters


def get_cluster(dc_ref, cluster):
    """
    Returns a cluster in a datacenter.

    dc_ref
        The datacenter reference

    cluster
        The cluster to be retrieved
    """
    dc_name = utils_common.get_managed_object_name(dc_ref)
    log.trace("Retrieving cluster '%s' from datacenter '%s'", cluster, dc_name)
    si = utils_common.get_service_instance_from_managed_object(dc_ref, name=dc_name)
    traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
        path="hostFolder",
        skip=True,
        type=vim.Datacenter,
        selectSet=[
            vmodl.query.PropertyCollector.TraversalSpec(
                path="childEntity", skip=False, type=vim.Folder
            )
        ],
    )
    items = [
        i["object"]
        for i in utils_common.get_mors_with_properties(
            si,
            vim.ClusterComputeResource,
            container_ref=dc_ref,
            property_list=["name"],
            traversal_spec=traversal_spec,
        )
        if i["name"] == cluster
    ]
    if not items:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "Cluster '{}' was not found in datacenter " "'{}'".format(cluster, dc_name)
        )
    return items[0]


def create_cluster(dc_ref, cluster_name, cluster_spec):
    """
    Creates a cluster in a datacenter.

    dc_ref
        The parent datacenter reference.

    cluster_name
        The cluster name.

    cluster_spec
        The cluster spec (vim.ClusterConfigSpecEx).
        Defaults to None.
    """
    dc_name = utils_common.get_managed_object_name(dc_ref)
    log.trace("Creating cluster '%s' in datacenter '%s'", cluster_name, dc_name)
    try:
        dc_ref.hostFolder.CreateClusterEx(cluster_name, cluster_spec)
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)


def update_cluster(cluster_ref, cluster_spec):
    """
    Updates a cluster in a datacenter.

    cluster_ref
        The cluster reference.

    cluster_spec
        The cluster spec (vim.ClusterConfigSpecEx).
        Defaults to None.
    """
    cluster_name = utils_common.get_managed_object_name(cluster_ref)
    log.trace("Updating cluster '%s'", cluster_name)
    try:
        task = cluster_ref.ReconfigureComputeResource_Task(cluster_spec, modify=True)
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: " "{}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    utils_common.wait_for_task(task, cluster_name, "ClusterUpdateTask")


def delete_cluster(service_instance, cluster_name, datacenter_name):
    """
    Deletes a datacenter.

    service_instance
        The Service Instance Object

    cluster_name
        The name of the cluster to delete

    datacenter_name
        The datacenter name to which the cluster belongs
    """
    root_folder = utils_common.get_root_folder(service_instance)
    log.trace("Deleting cluster '%s' in '%s'", cluster_name, datacenter_name)
    try:
        dc_obj = utils_datacenter.get_datacenter(service_instance, datacenter_name)
        cluster_obj = get_cluster(dc_ref=dc_obj, cluster=cluster_name)
        task = cluster_obj.Destroy_Task()
    except vim.fault.NoPermission as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(
            "Not enough permissions. Required privilege: {}".format(exc.privilegeId)
        )
    except vim.fault.VimFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareApiError(exc.msg)
    except vmodl.RuntimeFault as exc:
        log.exception(exc)
        raise salt.exceptions.VMwareRuntimeError(exc.msg)
    utils_common.wait_for_task(task, cluster_name, "DeleteClusterTask")


def list_clusters(service_instance):
    """
    Returns a list of clusters associated with a given service instance.

    service_instance
        The Service Instance Object from which to obtain clusters.
    """
    return utils_common.list_objects(service_instance, vim.ClusterComputeResource)


def create_drs_rule(name, affinity, vm_refs, enabled, mandatory, cluster_ref):
    """
    Create a virtual machine to virtual machine affinity or anti affinity DRS rule

    name
        The name of the rule.

    affinity
        (boolean) Describes whether to make affinity or anti affinity rule.

    vm_refs
        Array of virtual machines associated with DRS rule.

    enabled
        (boolean) Enable the DRS rule being created.

    mandatory
        (boolean) Sets whether the rule being created is mandatory.

    cluster_ref
        Reference to cluster DRS rule is being created on.
    """
    if affinity:
        rule_spec = vim.cluster.AffinityRuleSpec()
    else:
        rule_spec = vim.cluster.AntiAffinityRuleSpec()
    rule_spec.vm = vm_refs
    rule_spec.enabled = enabled
    rule_spec.mandatory = mandatory
    rule_spec.name = name
    spec = vim.cluster.RuleSpec(info=rule_spec, operation="add")
    config_spec = vim.cluster.ConfigSpecEx(rulesSpec=[spec])
    task = cluster_ref.ReconfigureEx(config_spec, modify=True)
    utils_common.wait_for_task(task, "Cluster", "Create DRS rule Task")


def update_drs_rule(rule_ref, vm_refs, enabled, mandatory, cluster_ref):
    """
    Update a virtual machine to virtual machine affinity or anti affinity DRS rule

    rule_ref
        Reference to rule with same name.

    vm_refs
        Array of virtual machines associated with DRS rule.

    enabled
        (boolean) Enable the DRS rule being created. Defaults to True.

    mandatory
        (optional, boolean) Sets whether the rule being created is mandatory. Defaults to None.

    cluster_ref
        Reference to cluster DRS rule is being created on.
    """
    rule_ref.vm = vm_refs
    rule_ref.enabled = enabled
    rule_ref.mandatory = mandatory
    spec = vim.cluster.RuleSpec(info=rule_ref, operation="edit")
    config_spec = vim.cluster.ConfigSpecEx(rulesSpec=[spec])
    task = cluster_ref.ReconfigureEx(config_spec, modify=True)
    utils_common.wait_for_task(task, "Cluster", "Create DRS rule Task")


def drs_rule_info(rule):
    """
    Returns info on a DRS rule.

    rule
        Reference to DRS rule.
    """
    rule_info = {
        "name": rule.name,
        "uuid": rule.ruleUuid,
        "enabled": rule.enabled,
        "mandatory": rule.mandatory,
        "key": rule.key,
        "in_compliance": rule.inCompliance,
    }
    if type(rule) == vim.cluster.AntiAffinityRuleSpec or type(rule) == vim.cluster.AffinityRuleSpec:
        vms = []
        for vm in rule.vm:
            vms.append(vm.name)
        rule_info["type"] = "vm_affinity_rule"
        rule_info["vms"] = vms
        rule_info["affinity"] = check_affinity(rule)
    elif type(rule) == vim.cluster.VmHostRuleInfo:
        rule_info["type"] = "vm_host_rule"
        rule_info["vm_group_name"] = rule.vmGroupName
        rule_info["affine_host_group_name"] = rule.affineHostGroupName
        rule_info["anti_affine_host_group_name"] = rule.antiAffineHostGroupName
    elif type(rule) == vim.cluster.DependencyRuleInfo:
        rule_info["type"] = "dependency_rule"
        rule_info["vm_group"] = rule.vmGroup
        rule_info["depends_on_vm_group"] = rule.dependsOnVmGroup
    else:
        raise salt.exceptions.VMwareApiError({f"Unknown affinity rule type {type(rule)}"})
    return rule_info


def check_affinity(rule):
    """
    returns True if rule is Affine, or False if rule is AntiAffine.

    rule
        Reference to DRS rule.
    """
    if type(rule) == vim.cluster.AntiAffinityRuleSpec:
        return False
    elif type(rule) == vim.cluster.AffinityRuleSpec:
        return True
    else:
        raise Exception(f"Rule type {type(rule)} has no affinity.")



"""
Salt execution module for public IP
Provides methods to Create, Update, Read and Delete public IP.
"""
import json
import logging
import os

import requests
from requests.exceptions import HTTPError
from requests.exceptions import RequestException
from requests.exceptions import SSLError
from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request

log = logging.getLogger(__name__)

__virtualname__ = "vmc_public_ip"
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def list_(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
    cursor=None,
    page_size=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Retrieves public IPs from given SDDC.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_public_ip.list hostname=nsxt-manager.local  ...

    hostname
        The host name of NSX-T manager.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    authorization_host
        Hostname of the VMC cloud console.

    org_id
        The ID of organization to which the SDDC belongs to.

    sddc_id
        The ID of SDDC from which the public IPs should be retrieved.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page).

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    sort_by
        (Optional) Field by which records are sorted.

    sort_ascending
        (Optional) Boolean value to sort result in ascending order. Enabled by default.

    """
    log.info("Retrieving public IPs for SDDC %s", sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/public-ips"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id)

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_public_ip.list",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get(
    id,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves given public IP from the given SDDC.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_public_ip.get hostname=nsxt-manager.local id=public-ip-1 ...

    id
        public IP ID for which details should be retrieved.

    hostname
        The host name of NSX-T manager.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    authorization_host
        Hostname of the VMC cloud console.

    org_id
        The ID of organization to which the SDDC belongs to.

    sddc_id
        The ID of SDDC from which the public IP should be retrieved.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Retrieving Public IP %s for SDDC %s", id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/public-ips/{public_ip_id}"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, public_ip_id=id)

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_public_ip.get",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(
    id,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
):
    """
    Delete given public IP from given SDDC.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_public_ip.delete hostname=nsxt-manager.local id=public-ip-1 ...

    id
        ID of specific public IP to be deleted.

    hostname
        The host name of NSX-T manager.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    authorization_host
        Hostname of the VMC cloud console.

    org_id
        The ID of organization to which the SDDC belongs to.

    sddc_id
        The ID of SDDC from which public IP will be deleted.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Deleting Public IP %s for SDDC %s", id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/public-ips/{public_ip_id}"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, public_ip_id=id)
    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_public_ip.delete",
        responsebody_applicable=False,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def create(
    id,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
):
    """
    Create public IP for given SDDC.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_public_ip.create hostname=nsxt-manager.local id=public-ip-1 ...

    id
        The ID and name that the public IP will be created with.

    hostname
        The host name of NSX-T manager.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    authorization_host
        Hostname of the VMC cloud console.

    org_id
        The ID of organization to which the SDDC belongs to.

    sddc_id
        The ID of SDDC for which public IP belongs to.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    """
    log.info("Creating Public IP %s for SDDC %s", id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/public-ips/{public_ip_id}"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, public_ip_id=id)

    # not using json infrastructure as all fields are available from user_input
    data = {"ip": None, "display_name": id, "id": id}

    return vmc_request.call_api(
        method=vmc_constants.PUT_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_public_ip.create",
        data=data,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def update(
    id,
    display_name,
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    verify_ssl=True,
    cert=None,
):
    """
    Update public IP display name for given public IP ID for given SDDC.

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_public_ip.create hostname=nsxt-manager.local id=public-ip-1 display_name=public_ip ...

    id
        ID of the public IP to update.

    display_name
        The new name of the public IP.

    hostname
        The host name of NSX-T manager.

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations.

    authorization_host
        Hostname of the VMC cloud console.

    org_id
        The ID of organization to which the SDDC belongs to.

    sddc_id
        The ID of SDDC for which public IP belongs to.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    """
    log.info("Updating Public IP %s for SDDC %s", id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "cloud-service/api/v1/infra/public-ips/{public_ip_id}"
    )
    api_url = api_url.format(base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, public_ip_id=id)

    # not using json infrastructure as all feilds details are available from user_input
    data = {"display_name": display_name}

    return vmc_request.call_api(
        method=vmc_constants.PUT_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_public_ip.update",
        data=data,
        verify_ssl=verify_ssl,
        cert=cert,
    )



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

import salt.exceptions
import salt.modules.cmdmod
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import saltext.vmware.utils.common as utils_common

try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if PyVmomi is installed.
    """
    if HAS_PYVMOMI:
        return True

    return False, "Missing dependency: The salt.utils.vmware module requires pyVmomi."


# 3 datastore functions are still in utils/vmware.py:
# def list_datastores(service_instance):
# def get_datastore_files(service_instance, directory, datastores, container_object, browser_spec):
# def rename_datastore(datastore_ref, new_datastore_name):


def list_datastore_clusters(service_instance):
    """
    Returns a list of datastore clusters associated with a given service instance.

    service_instance
        The Service Instance Object from which to obtain datastore clusters.
    """
    return utils_common.list_objects(service_instance, vim.StoragePod)


def get_datastores_from_ref(
    service_instance,
    reference,
    datastore_names=None,
    backing_disk_ids=None,
    get_all_datastores=False,
):
    """
    Returns a list of vim.Datastore objects representing the datastores visible
    from a VMware object, filtered by their names, or the backing disk
    cannonical name or scsi_addresses

    service_instance
        The Service Instance Object from which to obtain datastores.

    reference
        The VMware object from which the datastores are visible.

    datastore_names
        The list of datastore names to be retrieved. Default value is None.

    backing_disk_ids
        The list of canonical names of the disks backing the datastores
        to be retrieved. Only supported if reference is a vim.HostSystem.
        Default value is None

    get_all_datastores
        Specifies whether to retrieve all disks in the host.
        Default value is False.
    """
    obj_name = utils_common.get_managed_object_name(reference)
    if get_all_datastores:
        log.trace("Retrieving all datastores visible to '%s'", obj_name)
    else:
        log.trace(
            "Retrieving datastores visible to '%s': names = (%s); " "backing disk ids = (%s)",
            obj_name,
            datastore_names,
            backing_disk_ids,
        )
        if backing_disk_ids and not isinstance(reference, vim.HostSystem):

            raise salt.exceptions.ArgumentValueError(
                "Unsupported reference type '{}' when backing disk filter "
                "is set".format(reference.__class__.__name__)
            )
    if (not get_all_datastores) and backing_disk_ids:
        # At this point we know the reference is a vim.HostSystem
        log.trace("Filtering datastores with backing disk ids: %s", backing_disk_ids)
        storage_system = get_storage_system(service_instance, reference, obj_name)
        props = utils_common.get_properties_of_managed_object(
            storage_system, ["fileSystemVolumeInfo.mountInfo"]
        )
        mount_infos = props.get("fileSystemVolumeInfo.mountInfo", [])
        disk_datastores = []
        # Non vmfs volumes aren't backed by a disk
        for vol in [i.volume for i in mount_infos if isinstance(i.volume, vim.HostVmfsVolume)]:

            if not [e for e in vol.extent if e.diskName in backing_disk_ids]:
                # Skip volume if it doesn't contain an extent with a
                # canonical name of interest
                continue
            log.trace(
                "Found datastore '%s' for disk id(s) '%s'",
                vol.name,
                [e.diskName for e in vol.extent],
            )
            disk_datastores.append(vol.name)
        log.trace("Datastore found for disk filter: %s", disk_datastores)
        if datastore_names:
            datastore_names.extend(disk_datastores)
        else:
            datastore_names = disk_datastores

    if (not get_all_datastores) and (not datastore_names):
        log.trace(
            "No datastore to be filtered after retrieving the datastores "
            "backed by the disk id(s) '%s'",
            backing_disk_ids,
        )
        return []

    log.trace("datastore_names = %s", datastore_names)

    # Use the default traversal spec
    if isinstance(reference, vim.HostSystem):
        # Create a different traversal spec for hosts because it looks like the
        # default doesn't retrieve the datastores
        traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
            name="host_datastore_traversal",
            path="datastore",
            skip=False,
            type=vim.HostSystem,
        )
    elif isinstance(reference, vim.ClusterComputeResource):
        # Traversal spec for clusters
        traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
            name="cluster_datastore_traversal",
            path="datastore",
            skip=False,
            type=vim.ClusterComputeResource,
        )
    elif isinstance(reference, vim.Datacenter):
        # Traversal spec for datacenter
        traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
            name="datacenter_datastore_traversal",
            path="datastore",
            skip=False,
            type=vim.Datacenter,
        )
    elif isinstance(reference, vim.StoragePod):
        # Traversal spec for datastore clusters
        traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
            name="datastore_cluster_traversal",
            path="childEntity",
            skip=False,
            type=vim.StoragePod,
        )
    elif (
        isinstance(reference, vim.Folder)
        and utils_common.get_managed_object_name(reference) == "Datacenters"
    ):
        # Traversal of root folder (doesn't support multiple levels of Folders)
        traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
            path="childEntity",
            selectSet=[
                vmodl.query.PropertyCollector.TraversalSpec(
                    path="datastore", skip=False, type=vim.Datacenter
                )
            ],
            skip=False,
            type=vim.Folder,
        )
    else:
        raise salt.exceptions.ArgumentValueError(
            "Unsupported reference type '{}'" "".format(reference.__class__.__name__)
        )

    items = utils_common.get_mors_with_properties(
        service_instance,
        object_type=vim.Datastore,
        property_list=["name"],
        container_ref=reference,
        traversal_spec=traversal_spec,
    )
    log.trace("Retrieved %s datastores", len(items))
    items = [i for i in items if get_all_datastores or i["name"] in datastore_names]
    log.trace("Filtered datastores: %s", [i["name"] for i in items])
    return [i["object"] for i in items]


def get_storage_system(service_instance, host_ref, hostname=None):
    """
    Returns a host's storage system
    """

    if not hostname:
        hostname = utils_common.get_managed_object_name(host_ref)

    traversal_spec = vmodl.query.PropertyCollector.TraversalSpec(
        path="configManager.storageSystem", type=vim.HostSystem, skip=False
    )
    objs = utils_common.get_mors_with_properties(
        service_instance,
        vim.HostStorageSystem,
        property_list=["systemFile"],
        container_ref=host_ref,
        traversal_spec=traversal_spec,
    )
    if not objs:
        raise salt.exceptions.VMwareObjectRetrievalError(
            "Host's '{}' storage system was not retrieved" "".format(hostname)
        )
    log.trace("[%s] Retrieved storage system", hostname)
    return objs[0]["object"]


def get_datastores(
    service_instance, datastore_name=None, datacenter_name=None, cluster_name=None, host_name=None
):
    """
    Gets datastores on the most specific of host_name, cluster_name, datacenter_name, or everywhere.

    Then optionally filters them by datastore_name.
    """
    if datacenter_name or cluster_name or host_name:
        reference = utils_common.find_filtered_object(
            service_instance,
            datacenter_name=datacenter_name,
            cluster_name=cluster_name,
            host_name=host_name,
        )
        return get_datastores_from_ref(
            service_instance,
            reference=reference,
            datastore_names=[datastore_name],
            get_all_datastores=not datastore_name,
        )
    else:
        # TODO: update get_datastores_from_ref to work recursively
        # get_datastores_from_ref doesn't actually find all datastores when
        #  searching everything, this should work
        datastores = utils_common.get_mors_with_properties(
            service_instance, vim.Datastore, property_list=["name"]
        )
        if not datastore_name:
            return [datastore["object"] for datastore in datastores]
        else:
            return [
                datastore["object"]
                for datastore in datastores
                if datastore["name"] == datastore_name
            ]


def enter_maintenance_mode(datastore_ref):
    """
    Put datastore in maintenance mode.

    datastore_ref
        Reference to datastore.
    """
    ret = datastore_ref.DatastoreEnterMaintenanceMode()
    utils_common.wait_for_task(ret.task, datastore_ref.name, "Put datastore in maintenance mode")
    if datastore_ref.summary.maintenanceMode == "inMaintenance":
        return True
    return False


def exit_maintenance_mode(datastore_ref):
    """
    Take datastore out of maintenance mode.

    datastore_ref
        Reference to datastore.
    """
    task = datastore_ref.DatastoreExitMaintenanceMode_Task()
    utils_common.wait_for_task(task, datastore_ref.name, "Take datastore out of maintenance mode")
    if datastore_ref.summary.maintenanceMode == "normal":
        return True
    return False



# SPDX-License-Identifier: Apache-2.0
"""
    :codeauthor: `Agnes Tevesz (agnes.tevesz@morganstanley.com)`

    salt.config.schemas.esxvm
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ESX Virtual Machine configuration schemas
"""
from salt.utils.schema import AnyOfItem
from salt.utils.schema import ArrayItem
from salt.utils.schema import BooleanItem
from salt.utils.schema import ComplexSchemaItem
from salt.utils.schema import DefinitionsSchema
from salt.utils.schema import IntegerItem
from salt.utils.schema import IPv4Item
from salt.utils.schema import NullItem
from salt.utils.schema import NumberItem
from salt.utils.schema import StringItem


class ESXVirtualMachineSerialBackingItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Serial Port Backing
    """

    title = "ESX Virtual Machine Serial Port Backing"
    description = "ESX virtual machine serial port backing"
    required = True

    uri = StringItem()
    direction = StringItem(enum=("client", "server"))
    filename = StringItem()


class ESXVirtualMachineDeviceConnectionItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Serial Port Connection
    """

    title = "ESX Virtual Machine Serial Port Connection"
    description = "ESX virtual machine serial port connection"
    required = True

    allow_guest_control = BooleanItem(default=True)
    start_connected = BooleanItem(default=True)


class ESXVirtualMachinePlacementSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Placement
    """

    title = "ESX Virtual Machine Placement Information"
    description = "ESX virtual machine placement property"
    required = True

    cluster = StringItem(
        title="Virtual Machine Cluster",
        description="Cluster of the virtual machine if it is placed to a cluster",
    )
    host = StringItem(
        title="Virtual Machine Host",
        description="Host of the virtual machine if it is placed to a standalone host",
    )
    resourcepool = StringItem(
        title="Virtual Machine Resource Pool",
        description="Resource pool of the virtual machine if it is placed to a resource pool",
    )
    folder = StringItem(
        title="Virtual Machine Folder",
        description="Folder of the virtual machine where it should be deployed, default is the datacenter vmFolder",
    )


class ESXVirtualMachineCdDriveClientSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine CD Drive Client
    """

    title = "ESX Virtual Machine Serial CD Client"
    description = "ESX virtual machine CD/DVD drive client properties"

    mode = StringItem(required=True, enum=("passthrough", "atapi"))


class ESXVirtualMachineCdDriveIsoSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine CD Drive ISO
    """

    title = "ESX Virtual Machine Serial CD ISO"
    description = "ESX virtual machine CD/DVD drive ISO properties"

    path = StringItem(required=True)


class ESXVirtualMachineCdDriveSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine CD Drives
    """

    title = "ESX Virtual Machine Serial CD"
    description = "ESX virtual machine CD/DVD drive properties"

    adapter = StringItem(
        title="Virtual Machine CD/DVD Adapter",
        description="Unique adapter name for virtual machine cd/dvd drive",
        required=True,
    )
    controller = StringItem(required=True)
    device_type = StringItem(
        title="Virtual Machine Device Type",
        description="CD/DVD drive of the virtual machine if it is placed to a cluster",
        required=True,
        default="client_device",
        enum=("datastore_iso_file", "client_device"),
    )
    client_device = ESXVirtualMachineCdDriveClientSchemaItem()
    datastore_iso_file = ESXVirtualMachineCdDriveIsoSchemaItem()
    connectable = ESXVirtualMachineDeviceConnectionItem()


class ESXVirtualMachineSerialSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Serial Port
    """

    title = "ESX Virtual Machine Serial Port Configuration"
    description = "ESX virtual machine serial port properties"

    type = StringItem(
        title="Virtual Machine Serial Port Type",
        required=True,
        enum=("network", "pipe", "file", "device"),
    )
    adapter = StringItem(
        title="Virtual Machine Serial Port Name",
        description="Unique adapter name for virtual machine serial port"
        "for creation an arbitrary value should be specified",
        required=True,
    )
    backing = ESXVirtualMachineSerialBackingItem()
    connectable = ESXVirtualMachineDeviceConnectionItem()
    yield_port = BooleanItem(
        title="Serial Port Yield", description="Serial port yield", default=False
    )


class ESXVirtualMachineScsiSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine SCSI Controller
    """

    title = "ESX Virtual Machine SCSI Controller Configuration"
    description = "ESX virtual machine scsi controller properties"
    required = True

    adapter = StringItem(
        title="Virtual Machine SCSI Controller Name",
        description="Unique SCSI controller name"
        "for creation an arbitrary value should be specified",
        required=True,
    )
    type = StringItem(
        title="Virtual Machine SCSI type",
        description="Type of the SCSI controller",
        required=True,
        enum=("lsilogic", "lsilogic_sas", "paravirtual", "buslogic"),
    )
    bus_sharing = StringItem(
        title="Virtual Machine SCSI bus sharing",
        description="Sharing type of the SCSI bus",
        required=True,
        enum=("virtual_sharing", "physical_sharing", "no_sharing"),
    )
    bus_number = NumberItem(
        title="Virtual Machine SCSI bus number",
        description="Unique bus number of the SCSI device",
        required=True,
    )


class ESXVirtualMachineSataSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine SATA Controller
    """

    title = "ESX Virtual Machine SATA Controller Configuration"
    description = "ESX virtual machine SATA controller properties"
    required = False
    adapter = StringItem(
        title="Virtual Machine SATA Controller Name",
        description="Unique SATA controller name"
        "for creation an arbitrary value should be specified",
        required=True,
    )
    bus_number = NumberItem(
        title="Virtual Machine SATA bus number",
        description="Unique bus number of the SATA device",
        required=True,
    )


class ESXVirtualMachineDiskSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Disk
    """

    title = "ESX Virtual Machine Disk Configuration"
    description = "ESX virtual machine disk properties"
    required = True

    size = NumberItem(title="Disk size", description="Size of the disk in GB", required=True)
    unit = StringItem(
        title="Disk size unit",
        description="Unit of the disk size, to VMware a " "GB is the same as GiB = 1024MiB",
        required=False,
        default="GB",
        enum=("KB", "MB", "GB"),
    )
    adapter = StringItem(
        title="Virtual Machine Adapter Name",
        description="Unique adapter name for virtual machine"
        "for creation an arbitrary value should be specified",
        required=True,
    )
    filename = StringItem(
        title="Virtual Machine Disk File",
        description="File name of the virtual machine vmdk",
    )
    datastore = StringItem(
        title="Virtual Machine Disk Datastore",
        description="Disk datastore where the virtual machine files will be placed",
        required=True,
    )
    address = StringItem(
        title="Virtual Machine SCSI Address",
        description="Address of the SCSI adapter for the virtual machine",
        pattern=r"\d:\d",
    )
    thin_provision = BooleanItem(
        title="Virtual Machine Disk Provision Type",
        description="Provision type of the disk",
        default=True,
        required=False,
    )
    eagerly_scrub = AnyOfItem(required=False, items=[BooleanItem(), NullItem()])
    controller = StringItem(
        title="Virtual Machine SCSI Adapter",
        description="Name of the SCSI adapter where the disk will be connected",
        required=True,
    )


class ESXVirtualMachineNicMapSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Nic Map
    """

    title = "ESX Virtual Machine Nic Configuration"
    description = "ESX Virtual Machine nic properties"
    required = False

    domain = StringItem()
    gateway = IPv4Item()
    ip_addr = IPv4Item()
    subnet_mask = IPv4Item()


class ESXVirtualMachineInterfaceSchemaItem(ComplexSchemaItem):
    """
    Configuration Schema Item for ESX Virtual Machine Network Interface
    """

    title = "ESX Virtual Machine Network Interface Configuration"
    description = "ESX Virtual Machine network adapter properties"
    required = True

    name = StringItem(
        title="Virtual Machine Port Group",
        description="Specifies the port group name for the virtual machine connection",
        required=True,
    )
    adapter = StringItem(
        title="Virtual Machine Network Adapter",
        description="Unique name of the network adapter, "
        "for creation an arbitrary value should be specified",
        required=True,
    )
    adapter_type = StringItem(
        title="Virtual Machine Adapter Type",
        description="Network adapter type of the virtual machine",
        required=True,
        enum=("vmxnet", "vmxnet2", "vmxnet3", "e1000", "e1000e"),
        default="vmxnet3",
    )
    switch_type = StringItem(
        title="Virtual Machine Switch Type",
        description="Specifies the type of the virtual switch for the virtual machine connection",
        required=True,
        default="standard",
        enum=("standard", "distributed"),
    )
    mac = StringItem(
        title="Virtual Machine MAC Address",
        description="Mac address of the virtual machine",
        required=False,
        pattern="^([0-9a-f]{1,2}[:]){5}([0-9a-f]{1,2})$",
    )
    mapping = ESXVirtualMachineNicMapSchemaItem()
    connectable = ESXVirtualMachineDeviceConnectionItem()


class ESXVirtualMachineMemorySchemaItem(ComplexSchemaItem):
    """
    Configurtation Schema Item for ESX Virtual Machine Memory
    """

    title = "ESX Virtual Machine Memory Configuration"
    description = "ESX Virtual Machine memory property"
    required = True

    size = IntegerItem(title="Memory size", description="Size of the memory", required=True)

    unit = StringItem(
        title="Memory unit",
        description="Unit of the memory, to VMware a " "GB is the same as GiB = 1024MiB",
        required=False,
        default="MB",
        enum=("MB", "GB"),
    )
    hotadd = BooleanItem(required=False, default=False)
    reservation_max = BooleanItem(required=False, default=False)


class ESXVirtualMachineCpuSchemaItem(ComplexSchemaItem):
    """
    Configurtation Schema Item for ESX Virtual Machine CPU
    """

    title = "ESX Virtual Machine Memory Configuration"
    description = "ESX Virtual Machine memory property"
    required = True

    count = IntegerItem(title="CPU core count", description="CPU core count", required=True)
    cores_per_socket = IntegerItem(
        title="CPU cores per socket",
        description="CPU cores per socket count",
        required=False,
    )
    nested = BooleanItem(
        title="Virtual Machine Nested Property",
        description="Nested virtualization support",
        default=False,
    )
    hotadd = BooleanItem(
        title="Virtual Machine CPU hot add", description="CPU hot add", default=False
    )
    hotremove = BooleanItem(
        title="Virtual Machine CPU hot remove",
        description="CPU hot remove",
        default=False,
    )


class ESXVirtualMachineConfigSchema(DefinitionsSchema):
    """
    Configuration Schema for ESX Virtual Machines
    """

    title = "ESX Virtual Machine Configuration Schema"
    description = "ESX Virtual Machine configuration schema"

    vm_name = StringItem(
        title="Virtual Machine name",
        description="Name of the virtual machine",
        required=True,
    )
    cpu = ESXVirtualMachineCpuSchemaItem()
    memory = ESXVirtualMachineMemorySchemaItem()
    image = StringItem(title="Virtual Machine guest OS", description="Guest OS type", required=True)
    version = StringItem(
        title="Virtual Machine hardware version",
        description="Container hardware version property",
        required=True,
    )
    interfaces = ArrayItem(
        items=ESXVirtualMachineInterfaceSchemaItem(),
        min_items=1,
        required=False,
        unique_items=True,
    )
    disks = ArrayItem(
        items=ESXVirtualMachineDiskSchemaItem(),
        min_items=1,
        required=False,
        unique_items=True,
    )
    scsi_devices = ArrayItem(
        items=ESXVirtualMachineScsiSchemaItem(),
        min_items=1,
        required=False,
        unique_items=True,
    )
    serial_ports = ArrayItem(
        items=ESXVirtualMachineSerialSchemaItem(),
        min_items=0,
        required=False,
        unique_items=True,
    )
    cd_dvd_drives = ArrayItem(
        items=ESXVirtualMachineCdDriveSchemaItem(),
        min_items=0,
        required=False,
        unique_items=True,
    )
    sata_controllers = ArrayItem(
        items=ESXVirtualMachineSataSchemaItem(),
        min_items=0,
        required=False,
        unique_items=True,
    )
    datacenter = StringItem(
        title="Virtual Machine Datacenter",
        description="Datacenter of the virtual machine",
        required=True,
    )
    datastore = StringItem(
        title="Virtual Machine Datastore",
        description="Datastore of the virtual machine",
        required=True,
    )
    placement = ESXVirtualMachinePlacementSchemaItem()
    template = BooleanItem(
        title="Virtual Machine Template",
        description="Template to create the virtual machine from",
        default=False,
    )
    tools = BooleanItem(
        title="Virtual Machine VMware Tools",
        description="Install VMware tools on the guest machine",
        default=False,
    )
    power_on = BooleanItem(
        title="Virtual Machine Power",
        description="Power on virtual machine afret creation",
        default=False,
    )
    deploy = BooleanItem(
        title="Virtual Machine Deploy Salt",
        description="Deploy salt after successful installation",
        default=False,
    )


class ESXVirtualMachineRemoveSchema(DefinitionsSchema):
    """
    Remove Schema for ESX Virtual Machines to delete or unregister virtual machines
    """

    name = StringItem(
        title="Virtual Machine name",
        description="Name of the virtual machine",
        required=True,
    )
    datacenter = StringItem(
        title="Virtual Machine Datacenter",
        description="Datacenter of the virtual machine",
        required=True,
    )
    placement = AnyOfItem(
        required=False, items=[ESXVirtualMachinePlacementSchemaItem(), NullItem()]
    )
    power_off = BooleanItem(
        title="Power off vm",
        description="Power off vm before delete operation",
        required=False,
    )


class ESXVirtualMachineDeleteSchema(ESXVirtualMachineRemoveSchema):
    """
    Deletion Schema for ESX Virtual Machines
    """


class ESXVirtualMachineUnregisterSchema(ESXVirtualMachineRemoveSchema):
    """
    Unregister Schema for ESX Virtual Machines
    """



"""
Manage VMware VMC SDDC
"""
import logging

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates
from saltext.vmware.utils import vmc_vcenter_request

log = logging.getLogger(__name__)

__virtualname__ = "vmc_sddc"
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def list_(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    include_deleted=False,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves list of SDDCs for the given organization

    Please refer the `VMC List All SDDCs documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc.get hostname=vmc.vmware.com  ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization from which SDDCs are retrieved

    include_deleted: Boolean
        (Optional) When ``True``, forces the result to also include deleted SDDCs.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Retrieving the List of SDDCs for the given organization %s", org_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs".format(base_url=api_base_url, org_id=org_id)

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["includeDeleted"],
        includeDeleted=include_deleted,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc.list_",
        params=params,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_by_id(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Returns a SDDC detail for the given SDDC Id

    Please refer the `VMC Get SDDC documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc.get_by_id hostname=vmc.vmware.com sddc_id=sddc_id...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        sddc_id of the SDDC which will be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
         If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Retrieving the SDDC details for the sddc %s in the organization %s", sddc_id, org_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs/{sddc_id}".format(
        base_url=api_base_url, org_id=org_id, sddc_id=sddc_id
    )
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc.get_by_id",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def create(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_name,
    num_hosts,
    provider,
    region,
    account_link_config=None,
    account_link_sddc_config=None,
    deployment_type=None,
    host_instance_type=None,
    msft_license_config=None,
    sddc_id=None,
    sddc_template_id=None,
    sddc_type=None,
    size=None,
    skip_creating_vxlan=False,
    sso_domain=None,
    storage_capacity=None,
    vpc_cidr=None,
    vxlan_subnet=None,
    validate_only=False,
    verify_ssl=True,
    cert=None,
):
    """
    Create a SDDC for given org

    Please refer the `VMC Create A New SDDC documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/post/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc.create hostname=vmc.vmware.com ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_name: String
        (Required) The name of SDDC that will be assigned to new created SDDC

    num_hosts: Integer As Int32
        (Required) Number of hosts in a SDDC

    provider: String
        (Required) Determines what additional properties are available based on cloud provider.
        Possible values are: AWS, ZEROCLOUD

    region: String
        (Required) AWS region where SDDC will be deployed

    account_link_config
        (Optional) The account linking configuration, we will keep this one and remove accountLinkSddcConfig finally.

        account_link_config expects to be passed as a dict having delay_account_link parameter
            delay_account_link: Boolean
            (Optional) Boolean flag identifying whether account linking should be delayed or not for the SDDC.

            .. code::

                {
                    "delay_account_link": false
                }

    account_link_sddc_config : List Of AccountLinkSddcConfig
        (Optional) A list of the SDDC linking configurations to use.

         AccountLinkSddcConfig has two parameters connected_account_id  and customer_subnet_ids
            connected_account_id:String
            (Optional) The ID of the customer connected account to work with.

            customer_subnet_ids: Array of String
            (Optional) List of subnet Ids

    deployment_type: String
        (Optional) Denotes if request is for a SingleAZ or a MultiAZ SDDC. Default is SingleAZ.
        Possible values are: SingleAZ, MultiAZ

    host_instance_type: String
        (Optional) The instance type for the hosts in the primary cluster of the SDDC.
        Possible values are: i3.metal, r5.metal, i3en.metal

    msft_license_config :
        (Optional) Indicates the desired licensing support, if any, of Microsoft software.
        It can be specified in the below format:

        .. code:: python

            "msft_license_config": [
                {
                    "academic_license": False,
                    "mssql_licensing": "DISABLED",
                    "windows_licensing": "DISABLED"

                }
            ]

        Possible values are:

        **academic_license**

        ``True`` if it is Academic Standard or ``False`` Commercial Standard License.

        **mssql_licensing**

        The status MSSQL licensing for this SDDC’s clusters. Possible values are: DISABLED, CUSTOMER_SUPPLIED, ENABLED

        **windows_licensing**

        The status of Windows licensing for this SDDC’s clusters. Possible values are: DISABLED, CUSTOMER_SUPPLIED, ENABLED
                                        Possible values are: DISABLED, CUSTOMER_SUPPLIED, ENABLED

        Please refer the `VMC Doc about msft_license_config <https://developer.vmware.com/apis/vmc/v1.1/data-structures/MsftLicensingConfig/>`_

    sddc_id: String As UUID
        (Optional) If provided, will be assigned as SDDC id of the provisioned SDDC.

    sddc_template_id : String As UUID
        (Optional) If provided, configuration from the template will applied to the provisioned SDDC.

    sddc_type: String
        (Optional) Denotes the sddc type, if the value is null or empty, the type is considered as default.

    size:String
        (Optional) The size of the vCenter and NSX appliances. “large” sddcSize corresponds to a ‘large’ vCenter appliance and ‘large’ NSX appliance. ‘medium’ sddcSize corresponds to ‘medium’ vCenter appliance and ‘medium’ NSX appliance. Value defaults to ‘medium’.
        Possible values are: nsx_small, medium, large, nsx_large

    skip_creating_vxlan : Boolean
        (Optional) skip creating vxlan for compute gateway for SDDC provisioning

    sso_domain : String
        (Optional) The SSO domain name to use for vSphere users. If not specified, vmc.local will be used.

    storage_capacity:  Integer As Int64
        (Optional) The storage capacity value to be requested for the sddc primary cluster, in GiBs. If provided, instead of using the direct-attached storage, a capacity value amount of separable storage will be used.

    vpc_cidr
        (Optional) AWS VPC IP range. Only prefix of 16 or 20 is currently supported. Example: 10.2.0.0/16, 10.2.32.0/20

    vxlan_subnet : String
        (Optional) VXLAN IP subnet in CIDR for compute gateway

    validate_only: Boolean
        (Optional) When ``True``, only validates the given sddc configuration without provisioning

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    For example:

        .. code::

            {
                    "account_link_config": {
                        "delay_account_link": false
                    },
                    "account_link_sddc_config": [
                        {
                            "connected_account_id": "string",
                            "customer_subnet_ids": [
                                "string"
                            ]
                        }
                    ],
                    "deployment_type": "SingleAZ",
                    "host_instance_type": "i3.metal",
                    "msft_license_config": {
                        "mssql_licensing": "string",
                        "windows_licensing": "string"
                    },
                    "sddc_name": "Salt-SDDC-1",
                    "num_hosts": 0,
                    "provider": "ZEROCLOUD",
                    "sddc_id": "string-UUID",
                    "sddc_template_id": "string",
                    "sddc_type": "OneNode",
                    "size": "medium",
                    "skip_creating_vxlan": false,
                    "sso_domain": "string",
                    "storage_capacity": 1,
                    "vpc_cidr": "string",
                    "vxlan_subnet": "string",
                    "region": "us-west-2"
            }
    """
    log.info("Creating a new SDDC %s in the organization %s", sddc_name, org_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs".format(base_url=api_base_url, org_id=org_id)

    allowed_dict = {
        "name": sddc_name,
        "num_hosts": num_hosts,
        "provider": provider,
        "region": region,
        "account_link_config": account_link_config,
        "account_link_sddc_config": account_link_sddc_config,
        "deployment_type": deployment_type,
        "host_instance_type": host_instance_type,
        "msft_license_config": msft_license_config,
        "sddc_id": sddc_id,
        "sddc_template_id": sddc_template_id,
        "sddc_type": sddc_type,
        "size": size,
        "skip_creating_vxlan": skip_creating_vxlan,
        "sso_domain": sso_domain,
        "storage_capacity": storage_capacity,
        "vpc_cidr": vpc_cidr,
        "vxlan_subnet": vxlan_subnet,
    }
    req_data = vmc_request._filter_kwargs(allowed_kwargs=allowed_dict.keys(), **allowed_dict)
    params = vmc_request._filter_kwargs(allowed_kwargs=["validateOnly"], validateOnly=validate_only)
    request_data = vmc_request.create_payload_for_request(vmc_templates.create_sddc, req_data)
    return vmc_request.call_api(
        method=vmc_constants.POST_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc.create",
        data=request_data,
        params=params,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    force_delete=False,
    retain_configuration=False,
    template_name=None,
    verify_ssl=True,
    cert=None,
):
    """
    Deletes the Given SDDC

    Please refer the `VMC Delete SDDC documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/delete/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc.delete hostname=vmc.vmware.com sddc_id=sddc_id ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        sddc_id which will be deleted

    force_delete: Boolean
        (Optional) If ``True``, will delete forcefully.
        Beware: do not use the force flag if there is a chance an active provisioning or deleting task is running against this SDDC. This option is restricted.

    retain_configuration: Boolean
        (Optional) If ``True``, the SDDC's configuration is retained as a template for later use.
        This flag is applicable only to SDDCs in ACTIVE state.

    template_name: String
        (Optional) Only applicable when retainConfiguration is also set to ``True``. When set, this value will be used as the name of the SDDC configuration template generated.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Deleting the given SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs/{sddc_id}".format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["force", "retain_configuration", "template_name"],
        force=force_delete,
        retain_configuration=retain_configuration,
        template_name=template_name,
    )

    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc.delete",
        params=params,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def update_name(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    sddc_new_name,
    verify_ssl=True,
    cert=None,
):
    """
    Updates name for Given SDDC

    Please refer the `VMC Patch SDDC documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/patch/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc.update_name hostname=vmc.vmware.com sddc_id ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        sddc_id for which name will update

    sddc_new_name
        The new name of the SDDC to be changed to

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Updating sddc name for the SDDC %s", sddc_id)
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs/{sddc_id}".format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id
    )

    data = {"name": sddc_new_name}
    return vmc_request.call_api(
        method=vmc_constants.PATCH_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc.update_name",
        data=data,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_vcenter_detail(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves vcenter Details from the Given SDDC

    Please refer the `VMC Get SDDC documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc.get_vcenter_detail hostname=vmc.vmware.com sddc_id ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        sddc_id from which vcenter details should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Retrieving the vCenter detail for the given SDDC %s", sddc_id)
    sddc_detail = get_by_id(
        hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl, cert
    )
    if "error" in sddc_detail:
        return sddc_detail
    vcenter_url = sddc_detail["resource_config"]["vc_url"]
    username = sddc_detail["resource_config"]["cloud_username"]
    password = sddc_detail["resource_config"]["cloud_password"]
    vcenter_detail = {
        "vcenter_url": vcenter_url,
        "username": username,
        "password": password,
    }
    result = {"description": "vmc_sddc.get_vcenter_detail", "vcenter_detail": vcenter_detail}
    return result


def get_vms(
    hostname,
    username,
    password,
    clusters=None,
    datacenters=None,
    folders=None,
    hosts=None,
    names=None,
    power_states=None,
    resource_pools=None,
    vms=None,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves the virtual machines from the SDDC vcenter

    Please refer the `vCenter Get VM List documentation <https://developer.vmware.com/apis/vsphere-automation/latest/vcenter/api/vcenter/vm/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vmc_sddc.get_vms hostname=vmc.vmware.com ...

    hostname
        The host name of vCenter console

    username
        username required to login to vCenter console

    password
        password required to login to vCenter console

    clusters: Array Of String
        (Optional) Clusters that must contain the virtual machine for the virtual machine to match the filter. If unset or empty, virtual machines in any cluster match the filter.
        When clients pass a value of this structure as a parameter, the field must contain identifiers for the resource type: ClusterComputeResource.
        When operations return a value of this structure as a result, the field will contain identifiers for the resource type: ClusterComputeResource.

    datacenters: Array Of String
        (Optional) datacenters that must contain the virtual machine for the virtual machine to match the filter.
         If unset or empty, virtual machines in any datacenter match the filter.
         When clients pass a value of this structure as a parameter, the field must contain identifiers for the resource type: Datacenter.
         When operations return a value of this structure as a result, the field will contain identifiers for the resource type: Datacenter.

    folders: Array Of String
        (Optional) Folders that must contain the virtual machine for the virtual machine to match the filter.
        If unset or empty, virtual machines in any folder match the filter.
        When clients pass a value of this structure as a parameter, the field must contain identifiers for the resource type: Folder.
        When operations return a value of this structure as a result, the field will contain identifiers for the resource type: Folder.

    hosts: Array Of String
        (Optional) Hosts that must contain the virtual machine for the virtual machine to match the filter.
        If unset or empty, virtual machines on any host match the filter.
        When clients pass a value of this structure as a parameter, the field must contain identifiers for the resource type: HostSystem.
        When operations return a value of this structure as a result, the field will contain identifiers for the resource type: HostSystem.

    names: Array Of String
        (Optional) Names that virtual machines must have to match the filter (see VM.Info.name).
        If unset or empty, virtual machines with any name match the filter.


    power_states: Array Of VmPowerState
        (Optional) Power states that a virtual machine must be in to match the filter
        The Power.State enumerated type defines the valid power states for a virtual machine.
        POWERED_OFF : The virtual machine is powered off.
        POWERED_ON : The virtual machine is powered on.
        SUSPENDED : The virtual machine is suspended.
        If unset or empty, virtual machines in any power state match the filter.

    resource_pools: Array Of String
        (Optional) Resource pools that must contain the virtual machine for the virtual machine to match the filter.
        If unset or empty, virtual machines in any resource pool match the filter.
        When clients pass a value of this structure as a parameter, the field must contain identifiers for the resource type: ResourcePool.
        When operations return a value of this structure as a result, the field will contain identifiers for the resource type: ResourcePool.

    vms: Array Of String
        (Optional) Identifiers of virtual machines that can match the filter.
        If unset or empty, virtual machines with any identifier match the filter.
        When clients pass a value of this structure as a parameter, the field must contain identifiers for the resource type: VirtualMachine.
        When operations return a value of this structure as a result, the field will contain identifiers for the resource type: VirtualMachine.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Retrieving the  virtual machines from the SDDC vCenter")
    api_url_base = vmc_request.set_base_url(hostname)
    api_url = "{base_url}api/vcenter/vm"
    api_url = api_url.format(base_url=api_url_base)
    allowed_dict = {
        "clusters": clusters,
        "datacenters": datacenters,
        "folders": folders,
        "hosts": hosts,
        "names": names,
        "power_states": power_states,
        "resource_pools": resource_pools,
        "vms": vms,
    }
    params = vmc_request._filter_kwargs(allowed_kwargs=allowed_dict.keys(), **allowed_dict)

    headers = vmc_vcenter_request.get_headers(hostname, username, password)
    return vmc_vcenter_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        headers=headers,
        description="vmc_sddc.get_vms",
        params=params,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def get_vms_by_sddc_id(
    hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None
):
    """
    Retrieves the virtual machines for the given SDDC.

    Please refer the `vCenter Get VM List documentation <https://developer.vmware.com/apis/vsphere-automation/latest/vcenter/api/vcenter/vm/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion id> vm_minion vmc_sddc.get_vms_by_sddc_id hostname=vmc.vmware.com sddc_id ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        Id of the SDDC for which VMs are retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Retrieving the virtual machines for the given SDDC %s", sddc_id)
    vcenter_detail_result = get_vcenter_detail(
        hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl, cert
    )
    if "error" in vcenter_detail_result:
        return vcenter_detail_result

    vcenter_detail = vcenter_detail_result["vcenter_detail"]
    vcenter_url = vcenter_detail["vcenter_url"]
    username = vcenter_detail["username"]
    password = vcenter_detail["password"]

    vcenter_hostname = vcenter_url[8 : len(vcenter_url) - 1]
    return get_vms(
        hostname=vcenter_hostname,
        username=username,
        password=password,
        verify_ssl=verify_ssl,
        cert=cert,
    )



# Copyright 2021 VMware, Inc.
# SPDX-License: Apache-2.0
import logging

import salt.exceptions
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.datastore as utils_datastore
import saltext.vmware.utils.esxi as utils_esxi
import saltext.vmware.utils.vsphere as utils_vmware

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim, VmomiSupport

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_datastore"


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def maintenance_mode(datastore_name, datacenter_name=None, service_instance=None, profile=None):
    """
    Put datastore in maintenance mode.

    datastore_name
        Name of datastore.

    datacenter_name
        (optional) Name of datacenter where datastore exists.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    assert isinstance(datastore_name, str)
    datastores = utils_datastore.get_datastores(
        service_instance, datastore_name=datastore_name, datacenter_name=datacenter_name
    )
    ds = datastores[0] if datastores else None
    ret = utils_datastore.enter_maintenance_mode(ds)
    if ret:
        return {"maintenanceMode": "inMaintenance"}
    return {"maintenanceMode": "failed to enter maintenance mode"}


def exit_maintenance_mode(
    datastore_name, datacenter_name=None, service_instance=None, profile=None
):
    """
    Take datastore out of maintenance mode.

    datastore_name
        Name of datastore.

    datacenter_name
        (optional) Name of datacenter where datastore exists.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    assert isinstance(datastore_name, str)
    datastores = utils_datastore.get_datastores(
        service_instance, datastore_name=datastore_name, datacenter_name=datacenter_name
    )
    ds = datastores[0] if datastores else None
    ret = utils_datastore.exit_maintenance_mode(ds)
    if ret:
        return {"maintenanceMode": "normal"}
    return {"maintenanceMode": "failed to exit maintenance mode"}


def get(
    datastore_name=None,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Return info about datastores.

    datastore_name
        Filter by this datastore name (required when cluster is not specified)

    datacenter_name
        Filter by this datacenter name (required when cluster is not specified)

    cluster_name
        Filter by this cluster name (required when datacenter is not specified)

    host_name
        Filter by this ESXi hostname (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    log.debug(f"Running {__virtualname__}.get")
    ret = []
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    datastores = utils_datastore.get_datastores(
        service_instance,
        datastore_name=datastore_name,
        datacenter_name=datacenter_name,
        cluster_name=cluster_name,
        host_name=host_name,
    )

    for datastore in datastores:
        summary = datastore.summary
        info = {
            "accessible": summary.accessible,
            "capacity": summary.capacity,
            "freeSpace": summary.freeSpace,
            "maintenanceMode": summary.maintenanceMode,
            "multipleHostAccess": summary.multipleHostAccess,
            "name": summary.name,
            "type": summary.type,
            "url": summary.url,
            "uncommitted": summary.uncommitted if summary.uncommitted else 0,
            "hosts": [host.key.name for host in datastore.host],
        }
        ret.append(info)

    return ret


def list_datastores(
    datastore_names=None,
    backing_disk_ids=None,
    backing_disk_scsi_addresses=None,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Returns a list of dict representations of the datastores visible to the
    proxy object. The list of datastores can be filtered by datastore names,
    backing disk ids (canonical names) or backing disk scsi addresses.
    Supported proxy types: esxi, esxcluster, esxdatacenter

    datastore_names
        List of the names of datastores to filter on

    backing_disk_ids
        List of canonical names of the backing disks of the datastores to filer.
        Default is None.

    backing_disk_scsi_addresses
        List of scsi addresses of the backing disks of the datastores to filter.
        Default is None.

    datacenter_name
        Filter by this datacenter name (required when cluster is not specified)

    cluster_name
        Filter by this cluster name (required when datacenter is not specified)

    host_name
        Filter by this ESXi hostname (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_datastore.list_datastores
        salt '*' vmware_datastore.list_datastores datastore_names=[ds1, ds2]
    """
    log.debug("Running vmware_datastore.list_datastores")
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    for host in hosts:
        # Default to getting all disks if no filtering is done
        get_all_datastores = (
            True
            if not (datastore_names or backing_disk_ids or backing_disk_scsi_addresses)
            else False
        )
        # Get the ids of the disks with the scsi addresses
        if backing_disk_scsi_addresses:
            log.debug("Retrieving disk ids for scsi addresses '%s'", backing_disk_scsi_addresses)
            disk_ids = [
                d.canonicalName
                for d in utils_common.get_disks(host, scsi_addresses=backing_disk_scsi_addresses)
            ]
            log.debug("Found disk ids '%s'", disk_ids)
            backing_disk_ids = backing_disk_ids.extend(disk_ids) if backing_disk_ids else disk_ids
        datastores = utils_datastore.get_datastores_from_ref(
            service_instance, host, datastore_names, backing_disk_ids, get_all_datastores
        )

        # Search for disk backed datastores if target is host
        # to be able to add the backing_disk_ids
        mount_infos = []
        if isinstance(host, vim.HostSystem):
            storage_system = utils_common.get_storage_system(service_instance, host, host.name)
            props = utils_common.get_properties_of_managed_object(
                storage_system, ["fileSystemVolumeInfo.mountInfo"]
            )
            mount_infos = props.get("fileSystemVolumeInfo.mountInfo", [])
        ret[host.name] = []
        for datastore in datastores:
            datastore_dict = {
                "name": datastore.name,
                "type": datastore.summary.type,
                "free_space": datastore.summary.freeSpace,
                "capacity": datastore.summary.capacity,
            }
            backing_disk_ids = []
            for vol in [
                i.volume
                for i in mount_infos
                if i.volume.name == datastore.name and isinstance(i.volume, vim.HostVmfsVolume)
            ]:

                backing_disk_ids.extend([e.diskName for e in vol.extent])
            if backing_disk_ids:
                datastore_dict["backing_disk_ids"] = backing_disk_ids
            ret[host.name].append(datastore_dict)
    return ret


def list_disk_partitions(
    disk_id=None,
    scsi_address=None,
    datacenter_name=None,
    cluster_name=None,
    host_name=None,
    service_instance=None,
    profile=None,
):
    """
    Lists the partitions on a disk.
    The disk can be specified either by the canonical name, or by the
    scsi_address.

    disk_id
        Canonical name of the disk.
        Either ``disk_id`` or ``scsi_address`` needs to be specified
        (``disk_id`` supersedes ``scsi_address``.

    scsi_address`
        Scsi address of the disk.
        ``disk_id`` or ``scsi_address`` needs to be specified
        (``disk_id`` supersedes ``scsi_address``.

    datacenter_name
        Filter by this datacenter name (required when cluster is not specified)

    cluster_name
        Filter by this cluster name (required when datacenter is not specified)

    host_name
        Filter by this ESXi hostname (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_datastore.list_disk_partitions scsi_address='vmhaba0:C0:T0:L0'
        salt '*' vmware_datastore.list_disk_partitions disk_id='naa.000000000000001'
    """
    log.debug("Running vmware_datastore.list_disk_partitions")
    ret = {}
    if not disk_id and not scsi_address:
        raise salt.exceptions.ArgumentValueError(
            "Either 'disk_id' or 'scsi_address' needs to be specified"
        )
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    for host in hosts:
        if not disk_id:
            scsi_address_to_lun = utils_common.get_scsi_address_to_lun_map(host)
            if scsi_address not in scsi_address_to_lun:
                raise salt.exceptions.VMwareObjectRetrievalError(
                    "Scsi lun with address '{}' was not found on host '{}'".format(
                        scsi_address, host.name
                    )
                )
            disk_id = scsi_address_to_lun[scsi_address].canonicalName
            log.trace(
                "[%s] Got disk id '%s' for scsi address '%s'",
                host.name,
                disk_id,
                scsi_address,
            )
        log.trace("Listing disk partitions on disk '%s' in host '%s'", disk_id, host.name)
        partition_info = utils_vmware.get_disk_partition_info(host, disk_id)
        ret[host.name] = []
        # NOTE: 1. The layout view has an extra 'None' partition for free space
        #       2. The orders in the layout/partition views are not the same
        for part_spec in partition_info.spec.partition:
            part_layout = [
                p for p in partition_info.layout.partition if p.partition == part_spec.partition
            ][0]
            part_dict = {
                "hostname": host.name,
                "device": disk_id,
                "format": partition_info.spec.partitionFormat,
                "partition": part_spec.partition,
                "type": part_spec.type,
                "sectors": part_spec.endSector - part_spec.startSector + 1,
                "size_KB": (part_layout.end.block - part_layout.start.block + 1)
                * part_layout.start.blockSize
                / 1024,
            }
            ret[host.name].append(part_dict)
    return ret


def list_datastore_clusters(
    host_name=None,
    datacenter_name=None,
    cluster_name=None,
    service_instance=None,
    profile=None,
):
    """
    Returns a list of datastore clusters for the specified host.

    host_name
        Filter by this ESXi hostname (optional)

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_datastore.list_datastore_clusters
    """
    log.debug("Running vmware_datastore.list_datastore_clusters")
    ret = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    hosts = utils_esxi.get_hosts(
        service_instance=service_instance,
        host_names=[host_name] if host_name else None,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
        get_all_hosts=host_name is None,
    )
    return utils_datastore.list_datastore_clusters(service_instance)
